#include "DistributeVec.h"

extern struct opts Options ;

long DistributeVector(struct sparsematrix A, int *X, int dir) 
{
  register long k, t, tmp ;
  register int np, q, proc ;
  int P, **ProcList ;
  long *NrProcs ; 
  long *ProcHistogram, *sums, *PNrElts, *Ns, *Nr, *iv ;
  long lo, hi, min, Pmax, l = 0 ;
  long Com, ComIn, ComOut, NrElts, MaxIn, MaxOut, MaxInOut, MaxNrElts ;
  double ComSc, MaxInSc, MaxOutSc, MaxInOutSc ;

  if ( dir == ROW )
    l = A.n ;
  else
    if ( dir == COL )
      l = A.m ;
    else
      Abort("DistributeVector(): Direction unknown") ;

  /*## init: ##*/
  P = A.NrProcs ;
  proc = 0 ; 

  /*## Allocate processor lists: ##*/
  ProcList = (int **) malloc(l * sizeof(int *)) ;
  NrProcs = (long *) malloc(l * sizeof(long)) ;
  if ( ProcList == NULL || NrProcs == NULL )
    Abort("DistributeVector(): Not enough memory for vector arrays !") ;

  /*## Allocate processor arrays: ##*/
  ProcHistogram = (long *) malloc((P+1) * sizeof(long)) ;
  PNrElts = (long *) malloc(P * sizeof(long)) ;
  sums = (long *) malloc(P * sizeof(long)) ;
  Ns = (long *) malloc(P * sizeof(long)) ;
  Nr = (long *) malloc(P * sizeof(long)) ;
  if ( ProcHistogram == NULL || PNrElts == NULL || 
       sums == NULL || Ns == NULL || Nr == NULL )
    Abort("DistributeVector(): Not enough memory for processor arrays !") ;
  

  /*## Clear memory: ##*/
  for ( t = 0 ; t < l ; t++ )
    {
      NrProcs[t] = 0 ;
      ProcList[t] = NULL ;
    }
  for ( q = 0 ; q < P ; q++ )
    ProcHistogram[q] = sums[q] = PNrElts[q] = Ns[q] = Nr[q] = 0 ;
  ProcHistogram[P] = 0 ;
  min = 0 ;

  /*## STEP 1: 
    ##  Generate processor lists, histogram and initialize sums: ##*/
  Pmax = GenerateProcList(A, ProcList, NrProcs, dir) ;
  GenerateHistogram(ProcHistogram, ProcList, NrProcs, l, A.NrProcs) ;
  InitSums(sums, ProcList, NrProcs, l, A.NrProcs) ;

#ifdef INFO
  /*## Print histogram : ##*/
  if ( dir == ROW )
    printf("\nNr matrix columns owned by p processors"
           " (after matrix distribution):\n" ) ;
  else
    printf("\nNr matrix rows owned by p processors"
           " (after matrix distribution):\n" ) ;
  Com = 0 ;
  for ( q = 0 ; q <= Pmax ; q++ )
    {
      printf(" p=%d: %ld \n", q, ProcHistogram[q]) ;
      if (q >= 2)
        Com += ProcHistogram[q] * (q-1) ;
    }
  printf("Communication for vector"
         " %c (after matrix distribution):\n", dir == ROW ? 'v' : 'u') ;
  printf(" vol           : %ld \n", Com) ;
 
#endif

  /*## Sort NrProcs in decreasing order: ##*/
  iv = CSort(NrProcs, l, Pmax) ;


  /*## STEP 2: 
    ##  Assign vector components with one processor :##*/
  lo = l - ProcHistogram[0] - ProcHistogram[1] ;
  hi = l - ProcHistogram[0] ;
  for ( t = lo ; t < hi ; t++ )
    {      
      k = iv[t] ; 
      np = NrProcs[t] ;

      /*## Check Number of processors: ##*/
      if ( np != 1 )
	Abort("DistributeVector(): Internal error step 2: np != 1") ;

      /*## Assign X[k] to unique owner of component k: ##*/
      proc = ProcList[k][0] ;
      X[k] = proc ;
      PNrElts[proc]++ ;
    }


  /*## STEP 3:
    ##  Assign vector components with more than 2 processors: ##*/
  if ( P > 2 )
    {
      lo = 0 ;
      hi = l - ProcHistogram[0] - ProcHistogram[1] - ProcHistogram[2] ;

      if ( Options.VectorPartition_Step3 == Decreasing3 )
	for ( k = 0 ; k < (hi-lo) / 2 ; k++ )
	  {
	    tmp = iv[lo+k] ;
	    iv[lo+k] = iv[hi-k-1] ;
	    iv[hi-k-1] = tmp ;
	    tmp = NrProcs[lo+k] ;
	    NrProcs[lo+k] = NrProcs[hi-k-1] ;
	    NrProcs[hi-k-1] = tmp ;
	  }
      else
	if ( Options.VectorPartition_Step3 == Random3 )
	  PermuteLongLong(iv, NrProcs, lo, hi-1) ;


      for ( t = lo ; t < hi ; t++ )
	{
	  k = iv[t] ; 
	  np = NrProcs[t] ;	
	    
	  /*## Check Number of processors: ##*/
	  if ( np < 3 || np > P )	
	    Abort("DistributeVector(): Internal error step 3: "
		  "np < 3 or np > P") ;
	    
	  /*## Find a processor with the lowest sum: ##*/
	  proc = ProcList[k][0] ;
	  min = sums[proc] ;
	  for ( q = 1 ; q < np ; q++ )
	    if ( sums[ProcList[k][q]] < min )
	      {
		proc = ProcList[k][q] ;
		min = sums[proc] ;	      
	      }
	  
	  /*## If there is more than one processor having the minimum sum
	       we can look at a secondary objective like 
	       best balance lowest send.
	       We do not use secondary objectives here. ##*/
	  
	  /*## Assign X[k] to this processor: ##*/
	  AssignVectorComponent(X, k, proc, ProcList[k], np, PNrElts, 
				sums, Ns, Nr) ;	  
	}
    }	      


  /*## Step 4:
    ##  Assign vector components with two processors: ##*/
  if ( P > 1 )
    {
      lo = l - ProcHistogram[0] - ProcHistogram[1] - ProcHistogram[2] ;
      hi = l - ProcHistogram[0] - ProcHistogram[1] ;

      for ( t = lo ; t < hi ; t++ )
	{
	  k = iv[t] ; 
	  np = NrProcs[t] ;	
	  
	  /*## Check Number of processors: ##*/
	  if ( np != 2 )	
	    Abort("DistributeVector(): Internal error step 4: np != 2") ;
	  
	  /*## Find the match with the minimum Ns(sender) + Nr(receiver). ##
               This gives the least loaded send-receive direction. */
	  if ( Ns[ProcList[k][0]] + Nr[ProcList[k][1]] < 
	       Ns[ProcList[k][1]] + Nr[ProcList[k][0]] )
	    q = 0 ;
	  else
	    q = 1 ;
	  
	  proc = ProcList[k][q] ;

	  /*## Assign X[k] to this processor: ##*/
	  AssignVectorComponent(X, k, proc, ProcList[k], np, PNrElts, 
				sums, Ns, Nr) ;	  
	}
    }
  

  /*## Step 5:
    ##  Assign vector components with no processors: ##*/
  lo = l - ProcHistogram[0] ;
  hi = l ;
  for ( t = lo ; t < hi ; t++ )
    {
      k = iv[t] ; 
      np = NrProcs[t] ;	
      
      /*## Check Number of processors: ##*/
      if ( np != 0 )	
	Abort("DistributeVector(): Internal error step 5: np != 0") ;
      
      /*## Find a processor with the minimum number of vector components: ##*/
      proc = 0;
      min = PNrElts[0] ;
      for ( q = 1 ; q < P ; q++ )
	if ( PNrElts[q] < min )
	  {
	    proc = q ;
	    min = PNrElts[proc] ;
	  }
      
      /*## Assign X[k] to this processor: ##*/
      X[k] = proc ;
      PNrElts[proc]++ ;
    }  
  

  /*## Gather some statistics: ##*/
  ComIn = ComOut = NrElts = MaxIn = MaxOut = MaxInOut = MaxNrElts = 0 ;
  for ( q = 0 ; q < P ; q++ )
    {
      ComIn += Nr[q] ;
      ComOut += Ns[q] ;
      NrElts +=  PNrElts[q] ;

      if ( Nr[q] > MaxIn )
	MaxIn = Nr[q] ;

      if ( Ns[q] > MaxOut )
	MaxOut = Ns[q] ;

      if ( Nr[q] + Ns[q] > MaxInOut )
	MaxInOut = Nr[q] + Ns[q] ;

      if ( PNrElts[q] > MaxNrElts )
	MaxNrElts = PNrElts[q] ;
    }

  /*## Check if ComIn is equal to ComOut: ##*/
  if ( ComIn != ComOut )
    Abort("DistributeVector(): Internal error: ComIn != ComOut") ;

  /*## Check if total number of elements is equal to length: ##*/
  if ( NrElts != l )
    Abort("DistributeVector(): Internal error: NrElts != length") ;

  ComSc    = ((double) ComIn  / (double) l) ;
  MaxInSc  = ((double) MaxIn  / (double) l) ;
  MaxOutSc = ((double) MaxOut / (double) l) ;
  MaxInOutSc  = ((double) MaxInOut / (double) l) ;

#ifdef INFO
  printf("Communication for vector "
         "%c (after vector distribution):\n", dir == ROW ? 'v' : 'u') ;
  printf(" vol           : %ld (%g)\n", ComIn, ComSc) ;
  printf(" avg = vol/P   : %g (%g)\n", (double) ComIn / (double) P, ComSc / P) ;
  if ( dir == ROW )
    {
      printf(" max sent      : %ld (%g)\n", MaxOut, MaxOutSc) ;
      printf(" max recd      : %ld (%g)\n", MaxIn, MaxInSc) ; 
    }
  else
    { /* roles of sends and receives are interchanged */
      printf(" max sent      : %ld (%g)\n", MaxIn, MaxInSc) ; 
      printf(" max recd      : %ld (%g)\n", MaxOut, MaxOutSc) ;
    }
  printf(" max sent+recd : %ld (%g)\n", MaxInOut, MaxInOutSc) ;
  printf("Nr components for vector "
         "%c (after vector distribution):\n", dir == ROW ? 'v' : 'u') ;
  printf(" avg           : %g (%g) \n", (double) l / (double) P, 
         (double) 1.0 / (double) P) ;
  printf(" max           : %ld (%g)\n", MaxNrElts, 
         (double) MaxNrElts /(double) l ) ;
#endif

  /*## Cleanup memory: ##*/
  if ( NrProcs != NULL )
    free(NrProcs) ;
  
  for ( t = 0 ; t < l ; t++ )
    if ( ProcList[t] != NULL )
      free(ProcList[t]) ;
  if ( ProcList != NULL )
    free(ProcList) ;

  if ( iv != NULL )
    free(iv) ;

  /*## Cleanup processor arrays: ##*/
  if ( Nr != NULL )
    free(Nr) ;
  if ( Ns != NULL )
    free(Ns) ;
  if ( sums != NULL )
    free(sums) ;
  if ( PNrElts != NULL )
    free(PNrElts) ;
  if ( ProcHistogram != NULL )
    free(ProcHistogram) ;

  return(ComIn) ;
}




long DistributeVectorsEqual(struct sparsematrix A, int *U, int *V) 
{
  register long t, tmp, k ;
  register int np, q, proc ;
  int P, **ProcListU, **ProcListV ;
  int **ProcListUV ; /* intersection of the Processor list of U and V */
  long *NrProcsU, *NrProcsV ;
  long *NrProcsUV ;  /* number of processors in intersection */
  long PmaxU, PmaxV, PmaxUV, l, npU, npV ;

  long *ProcHistogramU, *sumsU, *PNrEltsU, *NsU, *NrU ;
  long *ProcHistogramV, *sumsV, *PNrEltsV, *NsV, *NrV ;
  long *ProcHistogramUV, *iuv ;

  long lo, hi, m1, m2, min ;
  long Com, ComInU, ComOutU, NrEltsU, MaxInU, MaxOutU, MaxInOutU, MaxNrEltsU ;
  long ComInV, ComOutV, NrEltsV, MaxInV, MaxOutV, MaxInOutV, MaxNrEltsV ;
  double ComSc, MaxInSc, MaxOutSc, MaxInOutSc ;

  /*## Check that the matrix is square: ##*/
  if ( A.m != A.n )
    Abort("DistributeVector(): U and V cannot be distributed the same way"
	  "because the matrix is not square") ;

  /*## init: ##*/
  l = A.m ;
  P = A.NrProcs ;
  proc = 0 ; 
  min = 0 ;

  /*## Allocate processor lists: ##*/
  ProcListU = (int **) malloc(l * sizeof(int *)) ;
  ProcListV = (int **) malloc(l * sizeof(int *)) ;
  ProcListUV = (int **) malloc(l * sizeof(int *)) ;
  NrProcsU = (long *) malloc(l * sizeof(long)) ;
  NrProcsV = (long *) malloc(l * sizeof(long)) ;
  NrProcsUV = (long *) malloc(l * sizeof(long)) ;
  if ( ProcListU == NULL || NrProcsU == NULL || 
       ProcListV == NULL || NrProcsV == NULL ||
       ProcListUV == NULL || NrProcsUV == NULL )
    Abort("DistributeVector(): Not enough memory for vector arrays") ;
  
  /*## Allocate processor arrays: ##*/
  ProcHistogramU = (long *) malloc((P+1) * sizeof(long)) ;
  ProcHistogramV = (long *) malloc((P+1) * sizeof(long)) ;
  ProcHistogramUV = (long *) malloc((P+1) * sizeof(long)) ;
  PNrEltsU = (long *) malloc(P * sizeof(long)) ;
  PNrEltsV = (long *) malloc(P * sizeof(long)) ;
  sumsU = (long *) malloc(P * sizeof(long)) ;
  sumsV = (long *) malloc(P * sizeof(long)) ;
  NsU = (long *) malloc(P * sizeof(long)) ;
  NsV = (long *) malloc(P * sizeof(long)) ;
  NrU = (long *) malloc(P * sizeof(long)) ;
  NrV = (long *) malloc(P * sizeof(long)) ;
  if ( ProcHistogramU == NULL || ProcHistogramV == NULL || 
       ProcHistogramUV == NULL || PNrEltsU == NULL || PNrEltsV == NULL || 
       sumsU == NULL || NsU == NULL || NrU == NULL ||
       sumsV == NULL || NsV == NULL || NrV == NULL )
    Abort("DistributeVector(): Not enough memory for processor arrays") ;

      
  /*## Clear memory: ##*/
  for ( t = 0 ; t < l ; t++ )
    {
      NrProcsU[t] = NrProcsV[t] = NrProcsUV[t] = 0 ;
      ProcListU[t] = ProcListV[t] = ProcListUV[t] = NULL ;
    }  
  for ( q = 0 ; q < P ; q++ )
    {
      ProcHistogramU[q] = sumsU[q] = PNrEltsU[q] = NsU[q] = NrU[q] = 0 ;
      ProcHistogramV[q] = sumsV[q] = PNrEltsV[q] = NsV[q] = NrV[q] = 0 ;
      ProcHistogramUV[q] = 0 ;
    }
  ProcHistogramU[P] = ProcHistogramV[P] = ProcHistogramUV[P] = 0 ;


  /*## STEP 1: 
    ##  Generate processor lists, histogram and initialize sums: ##*/
  PmaxU = GenerateProcList(A, ProcListU, NrProcsU, COL) ;
  PmaxV = GenerateProcList(A, ProcListV, NrProcsV, ROW) ;
  PmaxUV = CalculateIntersection(A, ProcListUV, NrProcsUV, ProcListU, 
				 NrProcsU, ProcListV, NrProcsV, A.m) ;
  
  GenerateHistogram(ProcHistogramU, ProcListU, NrProcsU, A.m, A.NrProcs) ;
  GenerateHistogram(ProcHistogramV, ProcListV, NrProcsV, A.n, A.NrProcs) ;
  GenerateHistogram(ProcHistogramUV, ProcListUV, NrProcsUV, A.n, A.NrProcs) ;

#ifdef INFO
  /*## Print histograms : ##*/
  printf("\nNr matrix columns owned by p processors"
         " (after matrix distribution):\n" ) ;
  Com = 0 ;
  for ( q = 0 ; q <= PmaxV ; q++ )
    {
      printf(" p=%d: %ld \n", q, ProcHistogramV[q]) ;
      if (q >= 2)
        Com += ProcHistogramV[q] * (q-1) ;
    }
  printf("Communication for vector v (after matrix distribution):\n") ;
  printf(" vol           : %ld \n", Com) ;
 
  printf("\nNr matrix rows owned by p processors"
         " (after matrix distribution):\n" ) ;
  Com = 0 ; 
  for ( q = 0 ; q <= PmaxU ; q++ ) 
    { 
      printf(" p=%d: %ld \n", q, ProcHistogramU[q]) ; 
      if (q >= 2) 
        Com += ProcHistogramU[q] * (q-1) ; 
    } 
  printf("Communication for vector u (after matrix distribution):\n") ;
  printf(" vol           :  %ld \n", Com) ; 

  printf("\nNr rows/columns with p processors in the intersection:\n" ) ;
  for ( q = 0 ; q <= PmaxUV ; q++ )
    printf(" p=%d: %ld \n", q, ProcHistogramUV[q]) ;
#endif

  InitSums(sumsU, ProcListU, NrProcsU, A.m, A.NrProcs) ;
  InitSums(sumsV, ProcListV, NrProcsV, A.n, A.NrProcs) ;

  /*## Sort NrProcs of the intersection in decreasing order: ##*/
  iuv = CSort(NrProcsUV, l, PmaxUV) ;


  /*## STEP 2: 
    ##  Assign vector components with one processor in the intersection of
    ##  the processor lists of u and v: ##*/ 
  lo = l - ProcHistogramUV[0] - ProcHistogramUV[1] ;
  hi = l - ProcHistogramUV[0] ;

  for ( t = lo ; t < hi ; t++ )
    {      
      k = iuv[t] ; 
      np = NrProcsUV[t] ;
      npU = NrProcsU[k] ;
      npV = NrProcsV[k] ;

      /*## Check Number of processors: ##*/
      if ( np != 1 )
	Abort("DistributeVectorsEqual(): Internal error step 2") ;

      proc = ProcListUV[k][0] ;

      /*## Assign U[k] to this processor: ##*/
      AssignVectorComponent(U, k, proc, ProcListU[k], npU, PNrEltsU, 
			    sumsU, NsU, NrU) ;	  
      
      /*## Assign V[k] to this processor: ##*/
      AssignVectorComponent(V, k, proc, ProcListV[k], npV, PNrEltsV, 
			    sumsV, NsV, NrV) ;
    }


  /*## STEP 3: 
    ##  Assign vector components with more than 1 processor in the 
    ##  intersection of the processor lists of u and v:##*/ 
  if ( P > 1 )
    {
      lo = 0 ;
      hi = l - ProcHistogramUV[0] - ProcHistogramUV[1] ;

      for ( t = lo ; t < hi ; t++ )
	{
	  k = iuv[t] ; 
	  np = NrProcsUV[t] ;
	  npU = NrProcsU[k] ;
	  npV = NrProcsV[k] ;
	    
	  /*## Check Number of processors: ##*/
	  if ( np < 2 || np > P )	
	    Abort("DistributeVectorsEqual(): Internal error step 3") ;
	    
	  /*## Find a processor with the lowest total sum: ##*/
	  proc = ProcListUV[k][0] ;
	  min = sumsU[proc] + sumsV[proc] ;
	  for ( q = 1 ; q < np ; q++ )
	    if ( sumsU[ProcListUV[k][q]] + sumsV[ProcListUV[k][q]] < min )
	      {
		proc = ProcListUV[k][q] ;
		min = sumsU[proc] + sumsV[proc] ;	      
	      }
	  
	  /*## If there is more than one processor having the minimum sum
	       we can look at a secondary objective but 
	       We do not use secondary objectives here. ##*/
	  
	  /*## Assign U[k] to this processor: ##*/
	  AssignVectorComponent(U, k, proc, ProcListU[k], npU, PNrEltsU, 
				sumsU, NsU, NrU) ;	  

	  /*## Assign V[k] to this processor: ##*/
	  AssignVectorComponent(V, k, proc, ProcListV[k], npV, PNrEltsV, 
				sumsV, NsV, NrV) ;	  
	}
    }	      


  /*## STEP 4: 
    ##  Assign vector components with no processors in the
        intersection of the processor lists of u and v: ##*/ 

  /*## First we sort the ProcessorList in such a way that: 
       all vector components with NrProcsU=0 XOR NrProcsV=0 are in
           iuv[lo] ... iuv[m1-1] 
       all vector components with NrProcsU!=0 AND NrProcsV!=0 are in
           iuv[m1] ... iuv[m2-1] 
       all vector components with NrProcsU=0 AND NrProcsV=0 are in
           iuv[m2] ... iuv[hi-1]  ##*/

  lo = m1 = l - ProcHistogramUV[0] ;
  hi = m2 = l ;

  t = m1;
  while ( t < m2 )
    {
      np = NrProcsUV[t] ;
      /*## Check Number of processors: ##*/
      if ( np != 0 )
	Abort("DistributeVectorsEqual(): Internal error step 4: np != 0") ;

      if ( NrProcsU[iuv[t]] == 0 && NrProcsV[iuv[t]] == 0 )
        { /* swap to right: */
          tmp = iuv[m2-1] ;
          iuv[m2-1] = iuv[t] ;
          iuv[t] = tmp ;
          m2-- ;
        }
      else
        if ( NrProcsU[iuv[t]] == 0 || NrProcsV[iuv[t]] == 0 )
	  { /* swap to left: */
	    tmp = iuv[m1] ;
	    iuv[m1] = iuv[t] ;
	    iuv[t] = tmp ;
	    m1++ ;
            t++ ;
	  }
	else
	  t++ ;
    }

  
  /*## Step 4A:
    ##  Assign vector components k with NrProcsU[k]=0 XOR NrProcsV[k]=0: ##*/

  for ( t = lo ; t < m1 ; t++ )
    {
      k = iuv[t] ; 
      np = NrProcsUV[t] ;
      npU = NrProcsU[k] ;
      npV = NrProcsV[k] ;

      /*## Check Number of processors: ##*/
      if ( np != 0 )	
	Abort("DistributeVectorsEqual(): Internal error step 4A") ;

      /*## Find a processor with the lowest total sum: ##*/
      if ( npU != 0 && npV == 0 )
	{
	  proc = ProcListU[k][0] ;
	  min = sumsU[proc] + sumsV[proc] ;	    
	  for ( q = 1 ; q < npU ; q++ )
	    if ( sumsU[ProcListU[k][q]] + sumsV[ProcListU[k][q]] < min )
	      {	    
		proc = ProcListU[k][q] ;
		min = sumsU[proc] + sumsV[proc] ;	    
	      }
	}
      else
	if ( npV != 0 && npU == 0 )
	  {	  
	    proc = ProcListV[k][0] ;
            min = sumsU[proc] + sumsV[proc] ;
	    for ( q = 1 ; q < npV ; q++ )
	      if ( sumsU[ProcListV[k][q]] + sumsV[ProcListV[k][q]] < min )
		{	    
		  proc = ProcListV[k][q] ;
		  min = sumsU[proc] + sumsV[proc] ;	    
		}
	  }
	else
	  Abort("DistributeVectorsEqual(): Internal error step 4A") ;
      
      /*## Assign U[k] to this processor: ##*/
      AssignVectorComponent(U, k, proc, ProcListU[k], npU, PNrEltsU, 
			    sumsU, NsU, NrU) ;	  
      
      /*## Assign V[k] to this processor: ##*/
      AssignVectorComponent(V, k, proc, ProcListV[k], npV, PNrEltsV, 
			    sumsV, NsV, NrV) ;
    }

  
  /*## Step 4B:
    ##  Assign vector components k with NrProcsU[k]!=0 and NrProcsV[k]!=0: ##*/

  for ( t = m1 ; t < m2 ; t++ )
    {
      k = iuv[t] ; 
      np = NrProcsUV[t] ;
      npU = NrProcsU[k] ;
      npV = NrProcsV[k] ;
      
      /*## Check Number of processors: ##*/
      if ( np != 0 || npU < 1 || npU > P || npV < 1 || npV > P )	
	Abort("DistributeVectorsEqual(): Internal error step 4B") ;
	          
      /*## Find a processor with the lowest total sum: ##*/
      proc = ProcListU[k][0] ;
      min = sumsU[proc] + sumsV[proc] ;
      for ( q = 1 ; q < npU ; q++ )
	if ( sumsU[ProcListU[k][q]] + sumsV[ProcListU[k][q]] < min )
	  {	    
	    proc = ProcListU[k][q] ;
	    min = sumsU[proc] + sumsV[proc] ;	    
	  }
      for ( q = 0 ; q < npV ; q++ )
	if ( sumsU[ProcListV[k][q]] + sumsV[ProcListV[k][q]] < min )
	  {
	    proc = ProcListV[k][q] ;
	    min = sumsU[proc] + sumsV[proc] ;
	  }
	
      /*## Assign U[k] to this processor: ##*/
      AssignVectorComponent(U, k, proc, ProcListU[k], npU, PNrEltsU, 
			    sumsU, NsU, NrU) ;	  
      
      /*## Assign V[k] to this processor: ##*/
      AssignVectorComponent(V, k, proc, ProcListV[k], npV, PNrEltsV, 
			    sumsV, NsV, NrV) ;
    }
  
  
  /*## Step 4C:
    ##  Assign vector components k with NrProcsU[k]=0 and NrProcsV[k]=0: ##*/
  for ( t = m2 ; t < hi ; t++ )
    {
      k = iuv[t] ; 
      np = NrProcsUV[t] ;
      npU = NrProcsU[k] ;
      npV = NrProcsV[k] ;
      
      /*## Check Number of processors: ##*/
      if ( np != 0 || npU != 0 || npV != 0 )	
	Abort("DistributeVectorsEqual(): Internal error step 4C") ;
      
      /*## Find a processor with the minimum number of vector components: ##*/
      proc = 0;
      min = PNrEltsU[0] ;
      for ( q = 1 ; q < P ; q++ )
	if ( PNrEltsU[q] < min )
	  {
	    proc = q ;
	    min = PNrEltsU[proc] ;
	  }
      
      /*## Assign U[k] to this processor: ##*/
      AssignVectorComponent(U, k, proc, ProcListU[k], npU, PNrEltsU, 
			    sumsU, NsU, NrU) ;	  
      
      /*## Assign V[k] to this processor: ##*/
      AssignVectorComponent(V, k, proc, ProcListV[k], npV, PNrEltsV, 
			    sumsV, NsV, NrV) ;
    }


  /*## Gather some statistics: ##*/
  ComInU = ComOutU = NrEltsU = MaxInU = MaxOutU = MaxInOutU = MaxNrEltsU = 0 ;
  ComInV = ComOutV = NrEltsV = MaxInV = MaxOutV = MaxInOutV = MaxNrEltsV = 0 ;
  for ( q = 0 ; q < P ; q++ )
    {
      ComInU += NrU[q] ;
      ComOutU += NsU[q] ;
      NrEltsU += PNrEltsU[q] ;

      ComInV += NrV[q] ;
      ComOutV += NsV[q] ;
      NrEltsV += PNrEltsV[q] ;

      if ( NrU[q] > MaxInU )
	MaxInU = NrU[q] ;

      if ( NrV[q] > MaxInV )
	MaxInV = NrV[q] ;

      if ( NsU[q] > MaxOutU )
	MaxOutU = NsU[q] ;

      if ( NsV[q] > MaxOutV )
	MaxOutV = NsV[q] ;

      if ( NrU[q] + NsU[q] > MaxInOutU )
	MaxInOutU = NrU[q] + NsU[q] ;

      if ( NrV[q] + NsV[q] > MaxInOutV )
	MaxInOutV = NrV[q] + NsV[q] ;

      if ( PNrEltsU[q] > MaxNrEltsU )
        MaxNrEltsU = PNrEltsU[q] ;

      if ( PNrEltsV[q] > MaxNrEltsV )
        MaxNrEltsV = PNrEltsV[q] ;

    }

  /*## Check if ComIn is equal to ComOut: ##*/
  if ( ComInU != ComOutU )
    Abort("DistributeVectorsEqual(): Internal error: ComInU != ComOutU") ;
  if ( ComInV != ComOutV )
    Abort("DistributeVectorsEqual(): Internal error: ComInV != ComOutV") ;

  /*## Check if total number of elements is equal to length: ##*/
  if ( NrEltsU != l || NrEltsV != l )
    Abort("DistributeVector(): Internal error: NrElts != length") ;


#ifdef INFO
  ComSc    = ((double) ComInV  / (double) l) ;
  MaxInSc  = ((double) MaxInV  / (double) l) ;
  MaxOutSc = ((double) MaxOutV / (double) l) ;
  MaxInOutSc  = ((double) MaxInOutV / (double) l) ;
  printf("\nCommunication for vector v (after vector distribution): \n") ; 
  printf(" vol           : %ld (%g)\n", ComInV, ComSc) ; 
  printf(" avg = vol/P   : %g (%g)\n", (double)ComInV / (double)P, ComSc / P) ;
  printf(" max sent      : %ld (%g)\n", MaxOutV, MaxOutSc) ;
  printf(" max recd      : %ld (%g)\n", MaxInV, MaxInSc) ; 
  printf(" max sent+recd : %ld (%g)\n", MaxInOutV, MaxInOutSc) ;

  ComSc    = ((double) ComInU  / (double) l) ;
  MaxInSc  = ((double) MaxInU  / (double) l) ;
  MaxOutSc = ((double) MaxOutU / (double) l) ;
  MaxInOutSc  = ((double) MaxInOutU / (double) l) ;
  printf("\nCommunication for vector u (after vector distribution): \n") ;
  printf(" vol           : %ld (%g)\n", ComInU, ComSc) ; 
  printf(" avg = vol/P   : %g (%g)\n", (double) ComInU / (double) P, ComSc / P) ;
  /* roles of sends and receives are interchanged  */
  printf(" max sent      : %ld (%g)\n", MaxInU, MaxInSc) ; 
  printf(" max recd      : %ld (%g)\n", MaxOutU, MaxOutSc) ; 
  printf(" max sent+recd : %ld (%g)\n", MaxInOutU, MaxInOutSc) ;
  printf("\nNr components for vectors v, u (after vector distribution): \n") ; 
  printf(" avg =         : %g \n", (double) l / (double) P) ;
  printf(" max =         : %ld \n", MaxNrEltsU) ;
#endif


  /*## Cleanup memory: ##*/
  for ( t = 0 ; t < l ; t++ )
    if ( NrProcsU[t] > 0 && ProcListU[t] != NULL )
      free(ProcListU[t]) ;
  if ( ProcListU != NULL )
    free(ProcListU) ;

  for ( t = 0 ; t < l ; t++ )
    if ( NrProcsV[t] > 0 && ProcListV[t] != NULL )
      free(ProcListV[t]) ;
  if ( ProcListV != NULL )
    free(ProcListV) ;

  for ( t = 0 ; t < l ; t++ )
    if ( NrProcsUV[t] > 0 && ProcListUV[t] != NULL )
      free(ProcListUV[t]) ;
  if ( ProcListUV != NULL )
    free(ProcListUV) ;

  if ( NrProcsU != NULL )
    free(NrProcsU) ;
  if ( NrProcsV != NULL )
    free(NrProcsV) ;
  if ( NrProcsUV != NULL )
    free(NrProcsUV) ;


  /*## Cleanup processor arrays: ##*/
  if ( NrU != NULL )
    free(NrU) ;
  if ( NsU != NULL )
    free(NsU) ;
  if ( PNrEltsU != NULL )
    free(PNrEltsU) ;
  if ( sumsU != NULL )
    free(sumsU) ;
  if ( ProcHistogramU != NULL )
    free(ProcHistogramU) ;

  if ( NrV != NULL )
    free(NrV) ;
  if ( NsV != NULL )
    free(NsV) ;
  if ( PNrEltsV != NULL )
    free(PNrEltsV) ;
  if ( sumsV != NULL )
    free(sumsV) ;
  if ( ProcHistogramV != NULL )
    free(ProcHistogramV) ;

  if ( ProcHistogramUV != NULL )
    free(ProcHistogramUV) ;
  if ( iuv != NULL )
    free(iuv) ;

  return(ComInU + ComInV) ;
}




int GenerateProcList(struct sparsematrix A, int **ProcList, long *NrProcs,
		     int dir)
{
  register long t, k, np ;
  int q, Pmax ;
  long *index = NULL ;

  if ( dir == COL )
    index = A.i ;
  else
    if ( dir == ROW )
      index = A.j ;
    else
      Abort("GenerateProcList(): unknown direction") ;
    
  for ( q = 0, Pmax = 0 ; q < A.NrProcs ; q++ )
    {
      for ( t = A.Pstart[q] ; t < A.Pstart[q+1] ; t++ )
	{
	  k = index[t] ;
	  np = NrProcs[k] ; /* Number of procs that own matrix row/column k */
	  
	  if ( np == 0 )
	    {
	      ProcList[k] = (int *) malloc(sizeof(int)) ;
	      ProcList[k][np] = q ;
	      NrProcs[k]++ ;
              if (Pmax == 0)
                Pmax = 1;
	    }
	  else
	    if ( ProcList[k][np-1] != q )
	      {
		ProcList[k] = (int *) realloc(ProcList[k], 
					      (np+1) * sizeof(int)) ;
		if ( ProcList[k] == NULL )
		  Abort("GenerateProcList(): Not enough memory for realloc"
			"of ProcList") ;
		
		ProcList[k][np] = q ;
		NrProcs[k]++ ;
		if ( NrProcs[k] > Pmax ) 
		  Pmax = NrProcs[k];
	      }
	}
    }  

  return(Pmax) ;
}


int CalculateIntersection(struct sparsematrix A, int **ProcListUV, 
			  long *NrProcsUV, int **ProcListU, long *NrProcsU, 
			  int **ProcListV, long *NrProcsV, long l)
/* Calculates for each component k the intersection of the 
   processor lists for U[k] and V[k]. 
   Each processor list must be in increasing order. */
{
  long t, np, npU, npV ;
  int iu, iv, Pu, Pv, Pmax ;

  Pmax = 0 ;

  for ( t = 0 ; t < l ; t++ )
    {
      npU = NrProcsU[t] ;
      npV = NrProcsV[t] ;

      iu = iv = 0 ;
      while ( iu < npU && iv < npV )
	{
	  np = NrProcsUV[t] ;
	  if ( (Pu = ProcListU[t][iu]) == (Pv = ProcListV[t][iv]) )
	    {
	      ProcListUV[t] = (int *) realloc(ProcListUV[t], 
					      (np+1) * sizeof(int)) ;
	      if ( ProcListUV[t] == NULL )
		Abort("CalculateIntersection(): Not enough memory for realloc"
		      "of ProcList") ;
	      
	      ProcListUV[t][np] = Pu ;
	      NrProcsUV[t]++ ;
	      if ( NrProcsUV[t] > Pmax ) 
		Pmax = NrProcsUV[t] ;
	      
	      iu++ ;
	      iv++ ;
	    }
	  else
	    if ( Pu < Pv )
	      iu++ ;
	    else
	      iv++ ;
	}      
    }

  return(Pmax) ;
}

void GenerateHistogram(long *ProcHistogram, int **ProcList, long *NrProcs,
		       long l, int P)
{
  register long t ;
  int q ;
  
  /*## Fill histogram: ##*/
  for ( t = 0 ; t < l ; t++ )
    ProcHistogram[NrProcs[t]]++ ;
  
  /*## Check histogram: ##*/
  for ( q = 0, t = 0 ; q <= P ; q++ )
    t += ProcHistogram[q] ;
  if ( t != l )
    Abort("GenerateHistogram(): Internal Error: sum != l") ;
}

void InitSums(long *Sums, int **ProcList, long *NrProcs, long l, int P)
/* Counts inevitable communication operations (sends or receives)
   that are known at the start of the vector distribution. 
   If a processor list contains more than one processor, each
   processor has to send or receive something. */
{
  register long t, np ;
  int q ;
  
  /*## Fill Sums: ##*/
  for ( t = 0 ; t < l ; t++ )
    {
      np = NrProcs[t] ;
      if ( np >= 2 )
	for ( q = 0 ; q < np ; q++ )
	  Sums[ProcList[t][q]]++ ;
    }
}



void AssignVectorComponent(int *X, long k, int proc, int *ProcList, 
				  int NrProcsInList, long *PNrElts, 
				  long *sums, long *Ns, long *Nr)
{
  int q ;

  X[k] = proc ;
  PNrElts[proc]++ ;

  if ( NrProcsInList == 0 )
    return;  /* no communication necessary */
  else
    if (NrProcsInList == 1)
      { 
        if ( proc == ProcList[0] )
          return;  /* no communication necessary */
        else 
          {
            Ns[proc]++ ;
            sums[proc]++ ;
            Nr[ProcList[0]]++ ;
            sums[ProcList[0]]++ ;
          }
      }
    else
      {
        /*## If the assigned processor is in the processor list,
             increase sums with NrProcs-2 else increase the sums with NrProcs
             and increase the sends of this processor: ##*/
        if ( BSearchInt(ProcList, 0, NrProcsInList-1, proc) != INT_MIN )
          {
	    sums[proc] += NrProcsInList-2 ;
            Ns[proc] += NrProcsInList-1 ;
          }
        else
          {
	    sums[proc] += NrProcsInList ;
            Ns[proc] += NrProcsInList ;
          }
      
        /*## Increase the receives of the other processors: ##*/
        for ( q = 0 ; q < NrProcsInList ; q++ )
	  if ( ProcList[q] != proc )
	    Nr[ProcList[q]]++ ;
      }
}


void WriteVectorDistribution(int *PX, long l, int P, char *filename)
{
  long t ;
  FILE *fp ;

  fp = OpenFile(filename, "w") ;
  
  fprintf(fp, "%ld %d\n", l, P) ;
  
  for ( t = 0 ; t < l ; t++ )
    fprintf(fp, "%ld %d\n", t+1, PX[t]+1) ;

  CloseFile(filename, fp) ;
  
}
