#include "HKLFM.h"

extern struct opts Options ;

void CreateInitialBalancedPartition(struct biparthypergraph *pHG) 
{
  int P ;
  register long t ;
  long *nnzV, *iv ;
  long left, right ;
  long maxvtxwgt ;

  /*## Allocate memory: ##*/
  nnzV = (long *) malloc(pHG->NrVertices * sizeof(long)) ;
  if ( nnzV == NULL )
    Abort("CreateInitialBalancedPartition(): Not enough memory for "
	  "temp array") ;

  /*## Copy the number of elements in each row to array : ##*/
  maxvtxwgt = pHG->V[0].vtxwgt ;
  for ( t = 0 ; t < pHG->NrVertices ; t++ )
    {
      nnzV[t] = pHG->V[t].vtxwgt ;
      if ( nnzV[t] > maxvtxwgt )
	maxvtxwgt = nnzV[t] ;
    }

  /*## Sort nnzV in decreasing order: ##*/
  iv = CSort(nnzV, pHG->NrVertices, maxvtxwgt) ;
      
  /*## Randomly permute the equal values : ##*/
  left = right = 0 ;
  while ( left < pHG->NrVertices )
    {
      right = left ;
      while ( right < pHG->NrVertices && nnzV[right] == nnzV[left] )
	right++ ;

      PermuteLong(iv, left, right-1) ;
      left = right ;
    }

  /*## Free memory: ##*/
  if ( pHG->NrVertices > 0 && nnzV != NULL )
    free(nnzV) ;
  
  /*## Assign Vertices to one of the Partitions in a greedy way: ##*/
  pHG->WeightP[0] = pHG->WeightP[1] = 0 ;
  for ( t = 0 ; t < pHG->NrVertices ; t++ )
    {
      if ( pHG->WeightP[0] < pHG->WeightP[1] )
	P = 0 ;
      else 
	if (  pHG->WeightP[1] < pHG->WeightP[0] )
	  P = 1 ;
	else
          {
	    P = (int) ( (2.0*rand()) / (RAND_MAX+1.0) ) ;
            if (P > 1)
              P = 1;
          }
      
      pHG->V[iv[t]].partition = P ;
      pHG->WeightP[P] += pHG->V[iv[t]].vtxwgt ;
    }
  
  /*## Free memory: ##*/
  if ( pHG->NrVertices > 0 && iv != NULL)
    free(iv) ;
}


void ComputeInitialGains(struct biparthypergraph *pHG)
{
  register long v, n, t ;
  int P ;
  long gain, tmp ;
  long nF ; /* number of vertices in the FROM partition of the net */
  long nT ; /* number of vertices in the TO partition of the net */
  
  /*## Sort the vertices for each net on partition number and set iStartP1: */
  for ( n = 0 ; n < pHG->NrNets ; n++ )
    {
      pHG->N[n].iStartP1 = pHG->N[n].iEnd ;
      t = pHG->N[n].iStartP0 ;

      while ( t < pHG->N[n].iStartP1 )
	{
	  if ( pHG->V[pHG->NetAdjncy[t]].partition == 1 )
	    { /*## Swap: NetAdjncy[t] <-> NetAdjncy[iStartP1-1] ##*/
	      tmp = pHG->NetAdjncy[--pHG->N[n].iStartP1] ;
	      pHG->NetAdjncy[pHG->N[n].iStartP1] = pHG->NetAdjncy[t] ;
	      pHG->NetAdjncy[t] = tmp ;
	    }
	  else
	    t++ ;
	}
    }

  /*## compute the initial gains and Weights: ##*/
  pHG->WeightP[0] = pHG->WeightP[1] = 0 ;
  for ( v = 0 ; v < pHG->NrVertices ; v++ )
    {
      P = pHG->V[v].partition ;
      pHG->WeightP[P] += pHG->V[v].vtxwgt ;

      /*## Calculate the vertex gain from the adjacent critical nets: ##*/
      for ( t = pHG->V[v].iStart, gain = 0 ; t < pHG->V[v].iEnd ; t++ )
	{
	  n = pHG->VtxAdjncy[t] ;

	  if ( P == 0 )
	    {
	      nF = pHG->N[n].iStartP1 - pHG->N[n].iStartP0 ;
	      nT = pHG->N[n].iEnd - pHG->N[n].iStartP1 ;
	    }
	  else
	    {
	      nF = pHG->N[n].iEnd - pHG->N[n].iStartP1 ;
	      nT = pHG->N[n].iStartP1 - pHG->N[n].iStartP0 ;
	    }	

	  if ( nF == 1 )
	    {
	      if ( Options.FM_WeightedHyperEdges == FMWgtEdgeYes )
		gain += pHG->EdgeWeight[t] ;
	      else
		gain++ ;
	    }
	  if ( nT == 0 )
	    {
	      if ( Options.FM_WeightedHyperEdges == FMWgtEdgeYes )
		gain -= pHG->EdgeWeight[t] ;
	      else
		gain-- ;
	    }
	}

      /*## Insert the gain in the correct gainbucket: ##*/
      pHG->V[v].GBentry = BucketInsert(&(pHG->GBVtx[P]), gain, v) ; 

    }

  /*## Calculate the initial Communication: ##*/
  pHG->CurComm = 0 ;
  for ( n = 0 ; n < pHG->NrNets ; n++ )
    {
      if ( pHG->N[n].iStartP1 > pHG->N[n].iStartP0 && 
	   pHG->N[n].iStartP1 < pHG->N[n].iEnd )
	pHG->CurComm++ ;
    }
}


void ClearMoveLog(struct biparthypergraph *pHG) 
{
  
  /* no need to clear the values, just reset the counters. */

  pHG->CurVtxLog = pHG->MinVtxLog = 0 ;
}


void MoveVertex(struct biparthypergraph *pHG, long v)
{
  int P = pHG->V[v].partition ;

  /*## Decrease the Weight for the old partition: ##*/
  pHG->WeightP[P]  -= pHG->V[v].vtxwgt ;

  /*## Increase the Weight for the new partition: ##*/
  pHG->WeightP[!P] += pHG->V[v].vtxwgt ;

  /*## Update partition number: ##*/
  pHG->V[v].partition = !P ;
    
  /*## Lock the vertex: ##*/
  pHG->V[v].GBentry = NULL ;
}


void UpdateGains(struct biparthypergraph *pHG, long v)
{
  int P ;
  register long t, tt ;
  long n, vadj, nw ;
  struct bucketentry *pE ;
  long nF ; /* number of vertices in the FROM partition of the net */
  long nT ; /* number of vertices in the TO partition of the net */

  P = pHG->V[v].partition ;

  /*## For all adjacent nets of this vertex: 
       if the net is critical: adjust the gain of the adjacent vertices. ##*/

  for ( t = pHG->V[v].iStart ; t < pHG->V[v].iEnd ; t++ )
    {
      n = pHG->VtxAdjncy[t] ;

      if ( P == 0 )
	{
	  nT = pHG->N[n].iStartP1 - pHG->N[n].iStartP0 ;
	  nF = pHG->N[n].iEnd - pHG->N[n].iStartP1 ;
	}
      else
	{
	  nT = pHG->N[n].iEnd - pHG->N[n].iStartP1 ;
	  nF = pHG->N[n].iStartP1 - pHG->N[n].iStartP0 ;
	}	

      /*## Check critical nets before the move: ##*/
      if ( nT == 0 )
	{ /*## Increase the gains of all free vertices adjacent to net n: ##*/
	  for ( tt = pHG->N[n].iStartP0 ; tt < pHG->N[n].iEnd ; tt++ )
	    {
	      vadj = pHG->NetAdjncy[tt] ;
	      pE = pHG->V[vadj].GBentry ;
	      
	      if ( pE != NULL )
		{
		  if ( Options.FM_WeightedHyperEdges == FMWgtEdgeYes )
		    nw = (long) (pE->bucket)->value + pHG->EdgeWeight[t] ;
		  else
		    nw = (long) (pE->bucket)->value + 1 ;
		  
                  if (nw != (pE->bucket)->value )
                    /* equality could happen for zero weights */
		    pHG->V[vadj].GBentry = 
		      BucketMove(&(pHG->GBVtx[pHG->V[vadj].partition]), pE, nw) ;
		}
	      
	    }
	  /*## Increase communication : */
	  pHG->CurComm++ ; 
	}
      if ( nT == 1 )
	{ /*## Decrease the gain of the vertex adjacent to net n if it is free: */
	  if ( P == 0 )
	    vadj = pHG->NetAdjncy[pHG->N[n].iStartP0] ;
	  else
	    vadj = pHG->NetAdjncy[pHG->N[n].iStartP1] ;

	  pE = pHG->V[vadj].GBentry ;
	      
	  if ( pE != NULL )
	    { 
	      if ( Options.FM_WeightedHyperEdges == FMWgtEdgeYes )
		nw = (long) (pE->bucket)->value - pHG->EdgeWeight[t] ;
	      else
		nw = (long) (pE->bucket)->value - 1 ;

              if (nw != (pE->bucket)->value ) 
	        pHG->V[vadj].GBentry = 
		  BucketMove(&(pHG->GBVtx[pHG->V[vadj].partition]), pE, nw) ;
	    }
	}

      /*## Change the net distribution to reflect the move: ##*/
      if ( P == 0 )
	{
	  /* find the index of v: */
	  tt = pHG->N[n].iStartP1 ; 
	  while ( pHG->NetAdjncy[tt] != v && tt < pHG->N[n].iEnd )
	    tt++ ;

	  /* swap iStartP1 <-> tt: */
	  pHG->NetAdjncy[tt] = pHG->NetAdjncy[pHG->N[n].iStartP1] ;
	  pHG->NetAdjncy[pHG->N[n].iStartP1] = v ;

	  /* increase the size of iStartP1: */
	  pHG->N[n].iStartP1++ ;
	  
	}	  
      else
	{
	  /* find the index of v: */
	  tt = pHG->N[n].iStartP0 ; 
	  while ( pHG->NetAdjncy[tt] != v && tt < pHG->N[n].iStartP1 )
	    tt++ ;

	  /* decrease the size of iStartP1: */
	  pHG->N[n].iStartP1-- ;

	  /* swap iStartP1-1 <-> v: */
	  pHG->NetAdjncy[tt] = pHG->NetAdjncy[pHG->N[n].iStartP1] ;
	  pHG->NetAdjncy[pHG->N[n].iStartP1] = v ;	  
	}	  
      nT++ ;
      nF-- ;

      /*## Check critical nets after the move: ##*/
      if ( nF == 0 )
	{ /*## Decrease the gains of all free vertices adjacent to net n: ##*/
	  for ( tt = pHG->N[n].iStartP0 ; tt < pHG->N[n].iEnd ; tt++ )
	    {
	      vadj = pHG->NetAdjncy[tt] ;
	      pE = pHG->V[vadj].GBentry ;

	      if ( pE != NULL )
		{
		  if ( Options.FM_WeightedHyperEdges == FMWgtEdgeYes )
		    nw = (long) (pE->bucket)->value - pHG->EdgeWeight[t] ;
		  else
		    nw = (long) (pE->bucket)->value - 1 ;

                  if (nw != (pE->bucket)->value ) 
		    pHG->V[vadj].GBentry = 
		      BucketMove(&(pHG->GBVtx[pHG->V[vadj].partition]), pE, nw) ;
		}
	    }
	  /*## Decrease comm: */
	  pHG->CurComm-- ; 
	}
      if ( nF == 1 )
	{ /*## Increase the gain of the vertex adjacent to net n if it is free: */
	  if ( P == 1 )
	    vadj = pHG->NetAdjncy[pHG->N[n].iStartP0] ;
	  else
	    vadj = pHG->NetAdjncy[pHG->N[n].iStartP1] ;

	  pE = pHG->V[vadj].GBentry ;
	      
	  if ( pE != NULL )
	    {
	      if ( Options.FM_WeightedHyperEdges == FMWgtEdgeYes )
		nw = (long) (pE->bucket)->value + pHG->EdgeWeight[t] ;
	      else
		nw = (long) (pE->bucket)->value + 1 ;

	      if (nw != (pE->bucket)->value )
	        pHG->V[vadj].GBentry = 
		  BucketMove(&(pHG->GBVtx[pHG->V[vadj].partition]), pE, nw) ;
	    }
	}	  
    }
}



void RunHKLFM(struct biparthypergraph *pHG, double eps)
{
  int ns ;

#ifdef INFO2
  printf("RunHKLFM\n") ;
  fflush(stdout) ;
#endif

  /* OptComm = lowest communication in all runs of HKLFM
     MinComm = lowest commmunication in one run of HKLFM */
  pHG->MinComm = LONG_MAX-1 ;
  pHG->OptComm = LONG_MAX ;

  for ( ns = 0 ; ns < Options.FM_InitPart_NrRestarts ; ns++ )
    {
      CreateInitialBalancedPartition(pHG) ;

      HKLFM(pHG, eps, Options.FM_InitPart_MaxNrLoops, 
	    Options.FM_InitPart_MaxNrNoGainMoves) ;
      
#ifdef INFO2
      printf(" %ld (Opt: %ld)\n", pHG->MinComm, pHG->OptComm) ;
#endif      
    }  

#ifdef INFO2
  printf(" Communication: %ld\n", pHG->OptComm) ;
#endif
}

void RefineHKLFM(struct biparthypergraph *pHG, double eps)
{
#ifdef INFO2
  printf("RefineHKLFM\n") ;
  fflush(stdout) ;
#endif

  pHG->MinComm = LONG_MAX-1 ;  
  pHG->OptComm = LONG_MAX ;  

  HKLFM(pHG, eps, Options.FM_Refine_MaxNrLoops, 
	Options.FM_Refine_MaxNrNoGainMoves) ;

#ifdef INFO2
  printf(" %ld (Opt: %ld)\n", pHG->MinComm, pHG->OptComm) ;
  printf(" Communication: %ld\n", pHG->OptComm) ;
#endif    
}


void HKLFM(struct biparthypergraph *pHG, double eps, int MaxNrLoops, 
	  int MaxNrNoGainMoves)
{
  int P, nl ;
  register long t ;
  long LastComm, NrNoGainIterations, gainP0, gainP1, MaxWeight ;
  long v, v0, v1 ;

  MaxWeight = floor( ((1.0 + eps) * (pHG->WeightP[0]+pHG->WeightP[1]))/2) ;

  LastComm = pHG->MinComm + 1 ;
  nl = 0 ;
  while ( nl++ < MaxNrLoops && LastComm != pHG->MinComm ) 
    {
      ClearGainBucket(&(pHG->GBVtx[0])) ;
      ClearGainBucket(&(pHG->GBVtx[1])) ;
      ClearMoveLog(pHG) ;
      
      ComputeInitialGains(pHG) ;

      if ( nl == 1 )
	{
	  pHG->MinComm = pHG->CurComm ;
	  if ( pHG->OptComm == LONG_MAX )
	    pHG->OptComm = pHG->CurComm ;
	}

      /* OptComm = lowest communication in all runs of HKLFM
         MinComm = lowest commmunication in one run of HKLFM 
         LastComm = lowest commmunication of previous iteration of this HKLFM run 
         CurComm = commmunication of current partitioning */
 
      LastComm = pHG->MinComm ;

#ifdef INFO2
      printf(" %ld", pHG->CurComm) ;
      fflush(stdout) ;
#endif
      
      NrNoGainIterations = 0 ;
      
      while ( pHG->GBVtx[0].NrBuckets + pHG->GBVtx[1].NrBuckets > 0 && 
	      NrNoGainIterations < MaxNrNoGainMoves )
	{
	  v0 = GainBucketGetMaxValVertexNr(&(pHG->GBVtx[0])) ;
	  v1 = GainBucketGetMaxValVertexNr(&(pHG->GBVtx[1])) ;
	  
	  if ( v0 == LONG_MIN && v1 == LONG_MIN )
	    Abort("HKLFM(): Internal error") ;

	  if ( v0 != LONG_MIN && v1 == LONG_MIN )
	    {
	      if ( pHG->WeightP[1] + pHG->V[v0].vtxwgt < MaxWeight )
		{
		  v = GainBucketExtractMax(&(pHG->GBVtx[0])) ;
		  MoveVertex(pHG, v) ;
		  UpdateGains(pHG, v) ; 		      
		  pHG->VtxMoveLog[pHG->CurVtxLog++] = v ;
		}
	      else
		{
		  v = GainBucketExtractMax(&(pHG->GBVtx[0])) ;
		  pHG->V[v].GBentry = NULL ;
		}
	    }
	  
	  if ( v1 != LONG_MIN && v0 == LONG_MIN )
	    {
	      if ( pHG->WeightP[0] + pHG->V[v1].vtxwgt < MaxWeight )
		{
		  v = GainBucketExtractMax(&(pHG->GBVtx[1])) ;
		  MoveVertex(pHG, v) ;
		  UpdateGains(pHG, v) ; 		      
		  pHG->VtxMoveLog[pHG->CurVtxLog++] = v ;
		}
	      else
		{
		  v = GainBucketExtractMax(&(pHG->GBVtx[1])) ;
		  pHG->V[v].GBentry = NULL ;
		}
	    }
	  
	  if ( v0 != LONG_MIN && v1 != LONG_MIN )
	    {
	      /* check best gain : */
	      gainP0 = gainP1 = LONG_MIN ;
	      
	      if ( pHG->WeightP[1] + pHG->V[v0].vtxwgt < MaxWeight ) 
		gainP0 = GainBucketGetMaxVal(&(pHG->GBVtx[0])) ;
	      if ( pHG->WeightP[0] + pHG->V[v1].vtxwgt < MaxWeight ) 
		gainP1 = GainBucketGetMaxVal(&(pHG->GBVtx[1])) ;
	      
	      if ( gainP0 == LONG_MIN && gainP1 == LONG_MIN )
		{
		  P = (int) ((2.0*rand())/(RAND_MAX+1.0)) ;
                  if (P > 1)
                    P = 1;
		  v = GainBucketExtractMax(&(pHG->GBVtx[P])) ;
		  pHG->V[v].GBentry = NULL ;
		}
	      else
		{
		  if( gainP0 > gainP1 )
		    P = 0 ;
		  else
		    if ( gainP1 > gainP0 )
		      P = 1 ;
		    else
                      {
		        P = (int) ((2.0*rand())/(RAND_MAX+1.0)) ;
                        if (P > 1)
                          P = 1;
                      }
		  
		  v = GainBucketExtractMax(&(pHG->GBVtx[P])) ;
		  MoveVertex(pHG, v) ;
		  UpdateGains(pHG, v) ; 		      
		  pHG->VtxMoveLog[pHG->CurVtxLog++] = v ;
		}
	    }		  
	  
	  /*## If current partition is the best yet: save it #*/
	  if ( pHG->CurComm <= pHG->MinComm )
	    {
	      pHG->MinComm = pHG->CurComm ;
	      pHG->MinVtxLog = pHG->CurVtxLog ;
	      NrNoGainIterations = 0 ;
	    }
	  else
	    NrNoGainIterations++ ;
	}
      
      /*## Restore best partition: */
      for ( t = pHG->CurVtxLog-1 ; t >= pHG->MinVtxLog ; t-- )
	{
          v = pHG->VtxMoveLog[t] ;
          MoveVertex(pHG, v) ;
	}

      /*## If this run has lowest overall comm, save it: ##*/
      if ( pHG->MinComm <= pHG->OptComm )
	{
	  pHG->OptComm = pHG->MinComm ;
	  for ( t = 0 ; t < pHG->NrVertices ; t++ )
	    pHG->OptPartVtx[t] = pHG->V[t].partition ;
	}
    }
}




void MoveVtxInNetAdjncy(struct biparthypergraph *pHG, long v)
/* Moves vertex to the back, so that it will not be 
   eligible for matching any more. This saves much time.
   istartP1 is abused to mark the end of the eligible candidates
   this function is only used during coarsening,
   so before any partitioning. */
{
  register long t ;
  long n, v2, iv, iv2 ;

  for ( t = pHG->V[v].iStart ; t < pHG->V[v].iEnd ; t++ )
    {
      n = pHG->VtxAdjncy[t] ; 

      /*## Find vertex v in adjacency list of net n: ##*/
      iv = pHG->N[n].iStartP0 ;
      while ( pHG->NetAdjncy[iv] != v )
	iv++ ;

      /*## Swap vertex:  v <-> Net[n].iStartP1  ##*/
      iv2 = --pHG->N[n].iStartP1 ;
      v2 = pHG->NetAdjncy[iv2] ;
      pHG->NetAdjncy[iv2] = v ;
      pHG->NetAdjncy[iv] = v2 ;
    }
}


void FindMatchRM(struct biparthypergraph *pHG, struct contraction *pC, long v,
                 struct bucketentry *pdummy)
{
  register long t, tt ;
  long v2, n, hvtxwgt ;

  /*## Add v to the match list: ##*/
  /* The vertex numbers of a group of matched vertices
     are stored consecutively in the array Match.
     The start of a group is given by Map.  */
  pC->Map[pC->NrMatches+1] = pC->Map[pC->NrMatches] ;
  pC->NrMatches++ ;
  pC->Match[pC->Map[pC->NrMatches]++] = v ;
  hvtxwgt = pHG->V[v].vtxwgt ;

  /* use the gainbucketentry to indicate if v has been matched or not.  
     pHG->V[v].GBentry = pdummy  indicates that vertex v has been matched. */
  pHG->V[v].GBentry = pdummy ;
  MoveVtxInNetAdjncy(pHG, v) ;

  /*## Search in the adjacent nets for an unmatched vertex: ##*/  
  for ( t = pHG->V[v].iStart ; t < pHG->V[v].iEnd ; t++ )
    {
      n = pHG->VtxAdjncy[t] ; 
      
      for ( tt = pHG->N[n].iStartP0 ; tt < pHG->N[n].iStartP1 ; tt++ )
	{
	  v2 = pHG->NetAdjncy[tt] ; 
	  
	  /*## Add v2 to the current match if:
	         - v2 is unmatched (this is guaranteed by the preceding
                   MoveVtxInNetAdjncy)
	         - number of vertices in current match < MaxNrVertices
	         - the weight of the hypervertex < MaxVtxWgt            ##*/
	  if ( pHG->V[v2].GBentry == pdummy )
	    Abort("FindMatchRM(): Internal error") ;

	  if ( hvtxwgt + pHG->V[v2].vtxwgt < pC->MaxVtxWgt )
	    {	      
	      if ( pC->Map[pC->NrMatches] - pC->Map[pC->NrMatches-1] <
		   pC->MaxNrVertices )
		{
		  pC->Match[pC->Map[pC->NrMatches]++] = v2 ;
		  pHG->V[v2].GBentry = pdummy ;
		  MoveVtxInNetAdjncy(pHG, v2) ;
		  hvtxwgt += pHG->V[v2].vtxwgt ;
		}
	      else
		return ;
	    }
	}
    }
}


void FindMatchIM(struct biparthypergraph *pHG, struct contraction *pC, long v,
                 struct bucketentry *pdummy)
{
  register long t, tt ;
  long v2, n, hvtxwgt, ew ;
  struct gainbucket GB ;
  struct bucketentry *pE ;
 
  /*## Add v to the match list: ##*/
  pC->Map[pC->NrMatches+1] = pC->Map[pC->NrMatches] ;
  pC->NrMatches++ ;
  pC->Match[pC->Map[pC->NrMatches]++] = v ;
  hvtxwgt = pHG->V[v].vtxwgt ;

  /* use the gainbucketentry to indicate if v has been matched or not.  
     pHG->V[v].GBentry = pdummy  indicates that vertex v has been matched. */
  pHG->V[v].GBentry = pdummy ;
  MoveVtxInNetAdjncy(pHG, v) ;

  /*## Initialize the gain bucket: ##*/
  GB.NrBuckets = 0 ;
  GB.Root = NULL ;

  /*## Fill the gain bucket: ##*/
  for ( t = pHG->V[v].iStart ; t < pHG->V[v].iEnd ; t++ )
    {
      n = pHG->VtxAdjncy[t] ;
      
      for ( tt = pHG->N[n].iStartP0 ; tt < pHG->N[n].iStartP1 ; tt++ )
	{
	  v2 = pHG->NetAdjncy[tt] ;	  	  
	  pE = pHG->V[v2].GBentry ;	  

	  if ( pE == pdummy || ( pE != NULL && pE->vtxnr != v2 ) )
            Abort("FindMatchIM(): Internal error") ;

	  if ( pE == NULL )
	    { /*## Add the vertex to the gainbucket: ##*/
	      if ( Options.Coarsening_WeightedHyperEdges == WgtEdgeYes )
		ew = pHG->EdgeWeight[t] + pHG->EdgeWeight[tt] ;
	      else
		ew = 1 ;

	      pHG->V[v2].GBentry = BucketInsert(&GB, ew, v2) ;

	    }
	  else
	    { /*## Move the vertex to the new gainbucket: ##*/
	      if ( Options.Coarsening_WeightedHyperEdges == WgtEdgeYes )
	        ew = (pE->bucket)->value + pHG->EdgeWeight[t] 
                                         + pHG->EdgeWeight[tt] ; 
	      else 
	        ew = (pE->bucket)->value + 1 ;
		
              if ( ew != (pE->bucket)->value ) 
	        pHG->V[v2].GBentry = BucketMove(&GB, pE, ew) ;
	    }
	}
    }

  /*## Empty the gain bucket: ##*/
  tt = 0 ;
  while ( GB.NrBuckets > 0 && tt < (pC->MaxNrVertices - 1) )
    {
      v2 = GainBucketExtractMax(&GB) ;

      if ( hvtxwgt + pHG->V[v2].vtxwgt < pC->MaxVtxWgt )
	{
	  pC->Match[pC->Map[pC->NrMatches]++] = v2 ;
	  hvtxwgt += pHG->V[v2].vtxwgt ;
	  pHG->V[v2].GBentry = pdummy ;
	  MoveVtxInNetAdjncy(pHG, v2) ;
	  tt++ ;
	}
      else
        pHG->V[v2].GBentry = NULL ;
    }


  while ( GB.NrBuckets > 0 )
    {
      v2 = GainBucketExtractMax(&GB) ;
      pHG->V[v2].GBentry = NULL ;
    }
}

struct contraction CoarsenGraph(struct biparthypergraph *pHG,
				struct biparthypergraph *pCHG)
{
  register long t, tt ;
  struct contraction C ; /* contains the matches and maps of the contraction */
  long *VtxWeight ; /* stores the weights of the vertices */
  long *ivRand ;    /* stores the random order in which vertices are visited */
  long *ivWeight ;  /* stores the weighted order in which vertices are visited */
  long *iAdjncy ; 
  long v, u, n, maxvtxwgt, totalvtxwgt, edgeindex ; 
  struct bucketentry dummy ;
  struct bucketentry *pdummy ;

  pdummy = &dummy;

  /*## Allocate memory for Match and Map: ##*/
  C.Match = (long *) malloc(pHG->NrVertices * sizeof(long)) ;
  C.Map = (long *) malloc((pHG->NrVertices+1) * sizeof(long)) ;
  if ( C.Match == NULL || C.Map == NULL )
    Abort("CoarsenGraph(): Not enough memory for Match / Map") ;

  ivRand = ivWeight = NULL ;

  /*## Init ivRand or ivWeight: ##*/
  totalvtxwgt = 0 ;
  switch (Options.Coarsening_MatchingStrategy)
    {
    case Random: 
      /*## Allocate memory for ivRand: ##*/
      ivRand = (long *) malloc(pHG->NrVertices * sizeof(long)) ;
      if ( ivRand == NULL )
	Abort("CoarsenGraph(): Not enough memory for ivRand") ;
      
      /*## Fill ivRand: ##*/
      for ( t = 0 ; t < pHG->NrVertices ; t++ )
	{
	  totalvtxwgt += pHG->V[t].vtxwgt ;
	  ivRand[t] = t ;
	  
	  /*## Permute the vertex adjacency lists: ##*/
	  if ( pHG->MatReValue == NULL )
            /* pattern matrix or numerical values not stored */
	    PermuteLong(pHG->VtxAdjncy, pHG->V[t].iStart, 
			pHG->V[t].iEnd-1) ;
	  else
            {
	      if ( pHG->MatImValue == NULL )
                /* real matrix with numerical values stored */
	        PermuteLongDouble(pHG->VtxAdjncy, pHG->MatReValue, 
		                  pHG->V[t].iStart, pHG->V[t].iEnd-1) ;
              else
                /* complex matrix with numerical values stored */
	        PermuteLongDoubleDouble(pHG->VtxAdjncy, pHG->MatReValue, 
		                  pHG->MatImValue, pHG->V[t].iStart,
                                  pHG->V[t].iEnd-1) ;
            }
	}
      
      /*## Permute ivRand to visit vertices randomly: ##*/
      PermuteLong(ivRand, 0, pHG->NrVertices-1) ;

      break ;

    case Inproduct: 
      /*## Allocate memory for VtxWeight: ##*/
      VtxWeight = (long *) malloc(pHG->NrVertices * sizeof(long)) ;
      if ( VtxWeight == NULL )
	Abort("CoarsenGraph(): Not enough memory for VtxWeight") ;    
      
      /*## Fill VtxWeight: ##*/
      maxvtxwgt = pHG->V[0].vtxwgt ;
      for ( t = 0 ; t < pHG->NrVertices ; t++ )
	{
	  totalvtxwgt += pHG->V[t].vtxwgt ;	  
	  VtxWeight[t] = pHG->V[t].vtxwgt ;

	  if ( VtxWeight[t] > maxvtxwgt )
	    maxvtxwgt = VtxWeight[t] ;
	}

      /*## Sort VtxWeight in decreasing order: ##*/
      ivWeight = CSort(VtxWeight, pHG->NrVertices, maxvtxwgt) ;	  

      /*## Free VtxWeight: ##*/
      if ( pHG->NrVertices > 0 && VtxWeight != NULL )
	free(VtxWeight) ;
      
      break ;

    default:
      Abort("CoarsenGraph(): Internal error: unknown matching strategy") ;
    }


  /*-------------------------------------------------------------------------
    ## Matching stage: ##
    -------------------------------------------------------------------------*/

  C.Map[0] = 0 ;
  C.NrMatches = 0 ;
  C.MaxNrVertices = Options.Coarsening_MaxNrVtxInMatch ;
  C.MaxVtxWgt = (long) floor(totalvtxwgt * 
			     Options.Coarsening_HVtxMaxFractionOfWeight) ;

  /*## Set iStartP1 equal to iEnd: ##*/
  for ( n = 0 ; n < pHG->NrNets ; n++ )
    pHG->N[n].iStartP1 = pHG->N[n].iEnd ;
  
  switch (Options.Coarsening_MatchingStrategy)
    {
    case Random: 
      for ( t = 0 ; t < pHG->NrVertices ; t++ )
	{
	  v = ivRand[t] ;	    
	  if ( pHG->V[v].GBentry != pdummy )
	    FindMatchRM(pHG, &C, v, pdummy) ; 
	}
      if ( pHG->NrVertices > 0 && ivRand != NULL )
	free(ivRand) ;

      break ;
    case Inproduct: 
      for ( t = 0 ; t < pHG->NrVertices ; t++ )
	{
	  v = ivWeight[t] ;
	  if ( pHG->V[v].GBentry != pdummy )
	    FindMatchIM(pHG, &C, v, pdummy) ; 
	}
      if ( pHG->NrVertices > 0 && ivWeight != NULL )
	free(ivWeight) ;

      break ;
    }

  /*## Reallocate the memory of C.Map to the number of matches: ##*/
  C.Map = (long *) realloc(C.Map, (C.NrMatches+1) * sizeof(long)) ;
  if ( C.Map == NULL )
    Abort("CoarsenGraph(): Not enough memory for realloc of C.Map") ;

  /*-------------------------------------------------------------------------
    ## Contraction stage: ##
    -------------------------------------------------------------------------*/
  *pCHG = CreateNewBiPartHyperGraph(C.NrMatches, pHG->NrNets, pHG->NrEdges, 
				    FALSE,'P') ;
   pCHG->SplitDir = pHG->SplitDir ;
  
  /*## Init temp array iAdjncy: ##*/
  iAdjncy = (long *) malloc(pCHG->NrNets * sizeof(long)) ;
  if ( iAdjncy == NULL )
    Abort("CoarsenGraph(): Not enough memory for iNetAjncy") ;
  /*## Clear iAdjncy: ##*/
  for ( t = 0 ; t < pCHG->NrNets ; t++ )
    iAdjncy[t] = -1 ;

  for ( u = 0, edgeindex = 0 ; u < C.NrMatches ; u++ )
    {
      pCHG->V[u].iStart = edgeindex ;
	  
      for ( t = C.Map[u] ; t < C.Map[u+1] ; t++ )
	{
	  v = C.Match[t] ;

	  /*## Add the weight of this vertex to the weight of u: ##*/
	  pCHG->V[u].vtxwgt += pHG->V[v].vtxwgt ;

	  /*## Add the nets of this vertex to the Adjacency list of u: ##*/
	  for ( tt = pHG->V[v].iStart ; tt < pHG->V[v].iEnd ; tt++ )
	    {
	      n = pHG->VtxAdjncy[tt] ;

	      if ( iAdjncy[n] == -1 )
		{ 
		  iAdjncy[n] = edgeindex ;
		  pCHG->VtxAdjncy[edgeindex++] = n ;

		  /*## Increase the counter for the net adjacency of n: ##*/
		  pCHG->N[n].iStartP1++ ;
		}

	      /*## Add the weight of edge tt to edge iAdjncy[n]: ##*/  
	      pCHG->EdgeWeight[iAdjncy[n]] += pHG->EdgeWeight[tt] ;
	    }
	}

      pCHG->V[u].iEnd = edgeindex ;

      /*## clear iAdjncy: ##*/
      for ( tt = pCHG->V[u].iStart ; tt < pCHG->V[u].iEnd ; tt++ )
	iAdjncy[pCHG->VtxAdjncy[tt]] = -1 ;
    }

  /*## Reset the NrEdges of the pCHG to the correct value: ##*/
  pCHG->NrEdges = edgeindex ;
  /*## Reallocate the memory to reflect this change: ##*/
  pCHG->VtxAdjncy = (long *) 
    realloc(pCHG->VtxAdjncy, pCHG->NrEdges * sizeof(long)) ;
  pCHG->NetAdjncy = (long *) 
    realloc(pCHG->NetAdjncy, pCHG->NrEdges * sizeof(long)) ;
  pCHG->EdgeWeight = (long *) 
    realloc(pCHG->EdgeWeight, pCHG->NrEdges * sizeof(long)) ;  
  if ( pCHG->VtxAdjncy == NULL || pCHG->NetAdjncy == NULL ||
       pCHG->EdgeWeight == NULL )
    Abort("CoarsenGraph(): Not enough memory for Reallocating NrEdges") ;


  /*## Set iStartP0, iStartP1 and iEnd for the nets of the CHG: ##*/
  /* iStartP1 is abused again */
  for ( t = 0, edgeindex = 0 ; t < pCHG->NrNets ; t++ )
    {
      /* N[t].istartP1 = number of adjacencies in net t */
      pCHG->N[t].iStartP0 = pCHG->N[t].iEnd = edgeindex ;
      edgeindex += pCHG->N[t].iStartP1 ;
      pCHG->N[t].iStartP1 = pCHG->N[t].iStartP0 ;
      /* N[t].istartP0 = N[t].istartP1 = N[t].iEnd
                       = start of adjacencies from net t */
    }

  /*## Set the net Adjacency for the CHG: ##*/
  for ( t = 0 ; t < pCHG->NrVertices ; t++ )
    for ( tt = pCHG->V[t].iStart ; tt < pCHG->V[t].iEnd ; tt++ )
      {
	n = pCHG->VtxAdjncy[tt] ;
	pCHG->NetAdjncy[pCHG->N[n].iEnd++] = t ;
      }

  /* for all t: N[t].iEnd = start of adjacencies from net t+1 */



  /*## Free memory: ##*/
  if ( iAdjncy != NULL )
    free(iAdjncy) ;

  return(C) ;
}

void RunMLGraphPart(struct biparthypergraph *pHG, double eps)
{
  long t, tt ;
  long nc ; /* number of contractions */
  struct biparthypergraph *CHG ; /* stores the sequence of coarser graphs */
  struct contraction *C ; /* stores the contraction information */
  double ratio ;
  
  /*## Coarsening phase: ##*/
  CHG = (struct biparthypergraph *) 
    malloc(sizeof(struct biparthypergraph)) ;
  if ( CHG == NULL )
    Abort("RunMLGraphPart(): Not enough memory for contracted hypergraph") ;

  C = NULL ;

  nc = 0 ;  
  CHG[nc] = *pHG ;

  while ( CHG[nc].NrVertices > Options.Coarsening_NrVertices )
    {
#ifdef INFO2
      printf("Coarsen CHG[%ld] (%ld) -> ", nc, 
	      CHG[nc].NrVertices) ;
#endif

      nc++ ;

      /*## Allocate memory: ##*/
      CHG = (struct biparthypergraph *) 
	realloc(CHG, (nc+1) * sizeof(struct biparthypergraph)) ;
      C = (struct contraction *) realloc(C, nc * sizeof(struct contraction)) ;
      if ( CHG == NULL || C == NULL )
	Abort("RunMLGraphPart(): Not enough memory for coarsening the Graph") ;

      /*## Coarsen the graph: ##*/
      C[nc-1] = CoarsenGraph(&CHG[nc-1], &CHG[nc]) ;

#ifdef INFO2
      printf("CHG[%ld] (%ld)\n", nc, CHG[nc].NrVertices) ;
#endif

      /*## Stop coarsening if number of matched vertices becomes small: ##*/
      ratio = ((double) CHG[nc-1].NrVertices - (double) CHG[nc].NrVertices) /
	(double) CHG[nc-1].NrVertices ;  

      if ( nc > 1 && ratio < Options.Coarsening_StopRatio )
	break ;

      if ( nc > 1 && CHG[nc].NrVertices == CHG[nc-1].NrVertices )
	break ;
    }

  /*## Initial partitioning: ##*/
#ifdef INFO2
  printf("Compute initial partition of CHG[%ld] \n", nc) ;
#endif
  RunHKLFM(&CHG[nc], eps) ;
#ifdef INFO2
  printf("WeightP[0] = %ld WeightP[1] = %ld\n", CHG[nc].WeightP[0],
	  CHG[nc].WeightP[1]) ;
#endif



  /*## Uncoarsening phase: ##*/
  while ( nc > 0 )
    {

#ifdef INFO2 
      printf("Copy partition info from CHG[%ld] to CHG[%ld]\n",
	      nc, nc-1) ;
#endif

      /*## Copy partition info : */
      for ( t = 0 ; t < C[nc-1].NrMatches ; t++ )
	for ( tt = C[nc-1].Map[t] ; tt < C[nc-1].Map[t+1] ; tt++ )
	  CHG[nc-1].V[C[nc-1].Match[tt]].partition = CHG[nc].OptPartVtx[t] ;
      
      CHG[nc-1].WeightP[0] = CHG[nc].WeightP[0] ;
      CHG[nc-1].WeightP[1] = CHG[nc].WeightP[1] ;
      
      /*## Delete CHG[nc] : */
      DeleteBiPartHyperGraph(CHG[nc]) ;

#ifdef INFO2 
      printf("Refine CHG[%ld]\n", nc-1) ;
#endif

      /*## Refinement phase: ##*/
      RefineHKLFM(&CHG[nc-1], eps) ;
      nc-- ;
    }

  *pHG = CHG[0] ;

#ifdef INFO2
  printf("Communication: %ld \n", pHG->OptComm) ;
#endif


  if ( CHG != NULL )
    free(CHG) ;
  if ( C != NULL )
    free(C) ;
}



