#include "Mondriaan.h"

/* input : matrix in  MatrixMarket coordinate format. */


/* Global variable:  */
struct opts Options ;


int main(int argc, char **argv) 
{
  int *u_proc ; /* Stores processor numbers for the vector components of u */
  int *v_proc ; /* Stores processor numbers for the vector components of v */
  double AvgNrEltsA ; /* Average number of nonzero elements per processor */
  long MinNrEltsA ; /* Minimum number of nonzero elements per processor */
  long MaxNrEltsA ; /* Maximum number of nonzero elements per processor */
  struct sparsematrix A ; /* The Matrix ;-) */
  long ComU, ComV ; 
  long q, nrelts ; 
  char *output ; 
  clock_t starttime, endtime ;
  double cputime;
#ifdef UNIX
  struct timeval starttime1, endtime1 ;
#endif


  /*## Get the parameters from the command line and Initialize Options: ##*/
  GetParameters(&Options, argc, argv) ; 

  /*## Allocate memory for filename: ##*/
  output = (char *) malloc((strlen(Options.matrix)+100) * sizeof(char)) ;
  if ( output == NULL )
    Abort("Not enough memory for output filename") ;    

  /*## Init; Read matrix: ##*/
  A = MMReadSparseMatrix(Options.matrix) ;

#ifdef INFO
  printf("\n**************************************************************\n") ;
  printf("Problem statistics:\n") ;
  printf(" Matrix:           : %s\n",Options.matrix) ;
  printf(" %s %s %s %s %s\n", A.Banner,A.Object,A.Format,A.Field,A.Symmetry) ;
  printf(" m = Nr rows       : %ld\n", A.m) ; 
  printf(" n = Nr columns    : %ld\n", A.n) ; 
  printf(" Nr matrix elements: %ld\n", A.NrNzElts) ;
#endif


  /*## Check if matrix A is already distributed: ##*/
  if ( A.MMTypeCode[0] == 'D' )
    {
      fprintf(stderr, "Warning: Matrix '%s' already distributed !\n", 
	      Options.matrix) ;
      fprintf(stderr, "         (Ignoring current partitions)\n") ; 
      
      if ( A.NrProcs != 0 && A.Pstart != NULL )
	free(A.Pstart) ;
    }

  /*## Symmetric options: ##*/
  if ( A.m == A.n && 
       (A.MMTypeCode[3]=='S' || A.MMTypeCode[3]=='K' || A.MMTypeCode[3]=='H'))
    {
      if ( Options.SymmetricMatrix_UseSingleEntry == SingleEntNo )
	SparseMatrixSymmetric2Full(&A) ;
      else
	if ( Options.SymmetricMatrix_SingleEntryType == ETypeRandom )
	  SparseMatrixSymmetricLower2Random(&A) ;
    }

  /*## If the matrix is square, add the dummies if requested: ##*/
  if ( A.m == A.n && 
       Options.SquareMatrix_DistributeVectorsEqual == EqVecYes &&
       Options.SquareMatrix_DistributeVectorsEqual_AddDummies == DumYes )
    AddDummiesToSparseMatrix(&A) ;

  /*## Set number of processors: ##*/
  A.NrProcs = Options.P ;
 
#ifdef INFO
  printf(" Nr processors     : %d\n", A.NrProcs) ;
  printf(" Nr dummies added  : %ld\n", A.NrDummies) ;
  printf(" allowed imbalance : %g %c\n", 100*Options.eps, '%') ;
  printf("Mondriaan matrix distribution:\n") ;
#endif

  /*## Init Pstart: ##*/
  A.Pstart = (long *) malloc((A.NrProcs+1) * sizeof(long)) ;
  if ( A.Pstart == NULL )
    Abort("Not enough memory for Pstart") ;
  for ( q = 0 ; q < A.NrProcs ; q++ )
    A.Pstart[q] = 0 ;
  A.Pstart[A.NrProcs] = A.NrNzElts ;

  /*## Distribute the matrix (and time it) ##*/
  starttime = clock() ;
#ifdef UNIX
  gettimeofday(&starttime1, NULL) ;
#endif

  DistributeMatrixMondriaan(&A, Options.P, Options.Alternate_FirstDirection, 
			    Options.eps) ; 
  endtime = clock() ;
#ifdef UNIX
  gettimeofday(&endtime1, NULL) ;
#endif
  cputime = ((double) (endtime - starttime)) / CLOCKS_PER_SEC ;

#ifdef TIME
  printf(" matrix distribution CPU-time    : %f seconds\n", cputime); 
#ifdef UNIX 
  printf(" matrix distribution elapsed time: %f seconds\n",
           (endtime1.tv_sec - starttime1.tv_sec) +
           (endtime1.tv_usec - starttime1.tv_usec) / 1000000.0 ) ; 
#endif 
  fflush(stdout) ;
#endif 

  /*## Remove the dummies: ##*/
  if ( A.m == A.n &&
       Options.SquareMatrix_DistributeVectorsEqual == EqVecYes && 
       Options.SquareMatrix_DistributeVectorsEqual_AddDummies == DumYes )
    RemoveDummiesFromSparseMatrix(&A) ;

  /*## Print information about the number of matrix elements: ##*/
  AvgNrEltsA = (double) A.NrNzElts / A.NrProcs ;
  MinNrEltsA = MaxNrEltsA = A.Pstart[1] ;
  for ( q = 0 ; q < A.NrProcs ; q++ )
    {
      nrelts = A.Pstart[q+1] - A.Pstart[q] ;
      if ( nrelts < MinNrEltsA )
	MinNrEltsA = nrelts ;
      if ( nrelts > MaxNrEltsA )
	MaxNrEltsA = nrelts ;
    }
#ifdef INFO
  printf(" Nr matrix elements:\n") ;
  printf("   tot         : %ld \n", A.NrNzElts) ;
  printf("   avg = tot/P : %g \n", AvgNrEltsA) ;
  printf("   max         : %ld \n", MaxNrEltsA) ;
  printf("   min         : %ld \n", MinNrEltsA) ;
  printf(" imbalance     : %g %c\n", 
	 100 * (MaxNrEltsA - AvgNrEltsA) / AvgNrEltsA, '%') ;
#endif

  /*## Convert randomly represented symmetric matrix to standard
       lower diagonal form: ##*/
  if ( A.m == A.n &&
       (A.MMTypeCode[3]=='S' || A.MMTypeCode[3]=='K' || A.MMTypeCode[3]=='H') &&
       Options.SymmetricMatrix_UseSingleEntry == SingleEntYes &&
       Options.SymmetricMatrix_SingleEntryType == ETypeRandom )
            SparseMatrixSymmetricRandom2Lower(&A) ;

  /*## Write the distributed matrix to file: ##*/
  sprintf(output, "%s-P%d", Options.matrix, Options.P) ;  
  MMWriteSparseMatrix(A, output) ;

  /*## Convert symmetric matrix to full form for vector distribution: ##*/
  if ( A.m == A.n &&
       (A.MMTypeCode[3]=='S' || A.MMTypeCode[3]=='K' || A.MMTypeCode[3]=='H') &&
       Options.SymmetricMatrix_UseSingleEntry == SingleEntYes )
         {
           SparseMatrixSymmetric2Full(&A) ;
           /*## Print information about the number of matrix elements: ##*/
           AvgNrEltsA = (double) A.NrNzElts / A.NrProcs ;
           MinNrEltsA = MaxNrEltsA = A.Pstart[1] ;
           for ( q = 0 ; q < A.NrProcs ; q++ )
             {
               nrelts = A.Pstart[q+1] - A.Pstart[q] ;
               if ( nrelts < MinNrEltsA )
                 MinNrEltsA = nrelts ;
               if ( nrelts > MaxNrEltsA )
                 MaxNrEltsA = nrelts ;
             }
#ifdef INFO
           printf(" After conversion of distributed symmetric matrix to"
                  " distributed full matrix\n" );
           printf(" Nr matrix elements:\n") ;
           printf("   tot         : %ld \n", A.NrNzElts) ;
           printf("   avg = tot/P : %g \n", AvgNrEltsA) ;
           printf("   max         : %ld \n", MaxNrEltsA) ;
           printf("   min         : %ld \n", MinNrEltsA) ;
           printf(" imbalance     : %g %c\n",
                  100 * (MaxNrEltsA - AvgNrEltsA) / AvgNrEltsA, '%') ;
#endif   
         }

  /* Vector Distribution: 
     ## Allocate memory for processor numbers for the vector components: ##*/
  u_proc = (int *) malloc(A.m * sizeof(int)) ;
  v_proc = (int *) malloc(A.n * sizeof(int)) ;
  if ( u_proc == NULL || v_proc == NULL )
    Abort("Not enough memory for vectors u_proc and v_proc") ;

  starttime = clock() ;
#ifdef UNIX
  gettimeofday(&starttime1, NULL) ;
#endif
 
  if ( A.m == A.n && Options.SquareMatrix_DistributeVectorsEqual == EqVecYes )
    {
      /*## Distribute the vectors equally: ##*/
      ComU = DistributeVectorsEqual(A, u_proc, v_proc) ;  
      ComV = 0 ;
    }
  else
    {
      /*## Distribute the vectors separately: ##*/
      ComV = DistributeVector(A, v_proc, ROW) ;
      ComU = DistributeVector(A, u_proc, COL) ;
    }

  endtime = clock() ;
#ifdef UNIX
  gettimeofday(&endtime1, NULL) ;
#endif
  cputime = ((double) (endtime - starttime)) / CLOCKS_PER_SEC ;

#ifdef TIME
  printf(" vector distribution CPU-time      : %f seconds\n", cputime);
#ifdef UNIX  
  printf(" vector distribution elapsed time  : %f seconds\n",
           (endtime1.tv_sec - starttime1.tv_sec) +
           (endtime1.tv_usec - starttime1.tv_usec) / 1000000.0 ) ; 
#endif   
  fflush(stdout) ;
#endif  

#ifdef INFO
  printf("Communication for u = A v: %ld (%g) \n", ComU + ComV,
	 (double) (ComU + ComV ) / (double) (A.m + A.n) ) ;  
  fflush(stdout) ;
#endif


  /*## Write vector elements to file: ##*/
  sprintf(output, "%s-u%d", Options.matrix, Options.P) ;    
  WriteVectorDistribution(u_proc, A.m, Options.P, output) ;

  sprintf(output, "%s-v%d", Options.matrix, Options.P) ;    
  WriteVectorDistribution(v_proc, A.n, Options.P, output) ;

  /*## Free memory: ##*/
  MMDeleteSparseMatrix(&A) ;  

  if ( A.m > 0 && u_proc != NULL )
    free(u_proc) ;
  if ( A.n > 0 && v_proc != NULL )
    free(v_proc) ;
  if ( output != NULL )
    free(output) ;
    
  /*## Exit :) ##*/
  return(0) ;
}



void GetParameters(struct opts *pOptions, int argc, char **argv)
{
  int t, i ;
  char c, *option, *value ;
  const char *eq = "=" ;

  /*## Print Help if NO option or a 'HELP' option is given: ##*/
  if ( argc == 1 || ! strcmp(argv[1], "--help") || ! strcmp(argv[1], "-h") ||
       ! strcmp(argv[1], "/?") || ! strcmp(argv[1], "-?") )
    {
      PrintHelp(argc, argv)  ;
      exit(1) ;
    }

  /*## Check the number of arguments: ##*/
  if ( argc < 4 )
    {
      printf("%s: invalid number of arguments \n", argv[0]) ;
      PrintHelp(argc, argv) ;
      exit(1) ;
    }    

  /*## Read the required arguments: ##*/
  pOptions->argc = argc ;
  pOptions->argv = argv ;
  pOptions->matrix = argv[1] ;
  pOptions->P = atoi(argv[2]) ;
  pOptions->eps = atof(argv[3]) ;

  /*## Set the default options: ##*/
  SetDefaults(pOptions) ;
  
  /*## Read the optional command line arguments: ##*/
  t = 4 ;
  while ( t < argc )
    {
      c = argv[t][0] ;
      option = argv[t] + 1 ;

      if ( strchr(option, '=') == NULL )
	{ 
	  value = argv[++t] ;	  

	  SetOption(pOptions, option, value) ;	  
	}
      else
	{
	  i = (int) strcspn(option, eq) ;

	  option[i] = '\0' ;
	  value = option + i + 1 ;

	  SetOption(pOptions, option, value) ;	  
	}  
    
      t++ ;
    }
}


void SetDefaults(struct opts *pOptions)
{
  FILE *fp ;
  int first, last ;
  unsigned int seed ;
  char filename[MAX_WORD_LENGTH] ;
  char *line, linebuffer[MAX_LINE_LENGTH] ;
  char option[MAX_WORD_LENGTH], value[MAX_WORD_LENGTH] ;
#ifdef UNIX
  struct timeval time ;
#endif

  strcpy(filename, pOptions->argv[0]) ;
  strcat(filename, ".defaults") ;

  /*## If the defaults file does not exist, create it: ##*/
  if ( (fp = fopen(filename, "r")) == NULL )
    {
      fprintf(stderr, "Cannot open file \"%s\".\n", filename) ;
      fprintf(stderr, "Creating file \"%s\" to store default options.\n", 
	      filename) ;

      /*## Write the options to file: ##*/
      fp = OpenFile(filename, "w") ;
      fprintf(fp, "# Mondriaan version %s default settings: \n", mversion) ;
      fprintf(fp, "LoadbalanceStrategy                            dynamic \n") ;
      fprintf(fp, "SplitStrategy                                  localbest \n");
      fprintf(fp, "Alternate_FirstDirection                       ratio \n") ;
      fprintf(fp, "SquareMatrix_DistributeVectorsEqual            no \n") ;
      fprintf(fp, "SquareMatrix_DistributeVectorsEqual_AddDummies yes \n") ;
      fprintf(fp, "SymmetricMatrix_UseSingleEntry                 no \n") ;
      fprintf(fp, "SymmetricMatrix_SingleEntryType                lower \n") ;
      fprintf(fp, "Coarsening_NrVertices                          200 \n") ;
      fprintf(fp, "Coarsening_MaxNrVtxInMatch                     2 \n") ;
      fprintf(fp, "Coarsening_StopRatio                           0.05 \n") ;
      fprintf(fp, "Coarsening_HVtxMaxFractionOfWeight             0.2 \n") ;
      fprintf(fp, "Coarsening_MatchingStrategy                    inproduct \n");
      fprintf(fp, "Coarsening_WeightedHyperEdges                  no \n") ;
      fprintf(fp, "FM_WeightedHyperEdges                          no \n") ;
      fprintf(fp, "FM_InitPart_NrRestarts                         8 \n") ;
      fprintf(fp, "FM_InitPart_MaxNrLoops                         25 \n") ;
      fprintf(fp, "FM_InitPart_MaxNrNoGainMoves                   200 \n") ;
      fprintf(fp, "FM_Refine_MaxNrLoops                           25 \n") ;
      fprintf(fp, "FM_Refine_MaxNrNoGainMoves                     200 \n") ;
      fprintf(fp, "VectorPartition_Step3                          random \n") ;
      fprintf(fp, "Seed                                           99 \n") ;
      CloseFile(filename, fp) ;
    }

  /*## Read the option lines from the file: ##*/
  fp = OpenFile(filename, "r") ;
  while ( (line = fgets(linebuffer, MAX_LINE_LENGTH, fp)) != NULL )
    { 
      first = 0 ; last = strlen(line) ;

      /*## Remove initial blank characters from the line: ##*/
      while ( isspace((int) linebuffer[first]) ) 
	first++ ;

      line = linebuffer + first ;

      /*## Clear option and value: ##*/
      option[0] = '\0' ;
      value[0]  = '\0' ;

      if ( first != last && linebuffer[first] != '#' )
	{ /*## Read option and value from the line: ##*/
	  sscanf(line, "%s %s", option, value) ;

	  /*## Set the option: ##*/
	  SetOption(pOptions, option, value) ;
	}
    }

  /*## Check option values: ##*/
  if ( pOptions->Coarsening_NrVertices < 1 )
    Abort("SetDefaults(): Coarsening_NrVertices out of range") ;

  if ( pOptions->Coarsening_MaxNrVtxInMatch < 2 )
    Abort("SetDefaults(): Coarsening_MaxNrVtxInMatch out of range") ;    

  if ( pOptions->Coarsening_StopRatio < 0 || 
       pOptions->Coarsening_StopRatio > 1 )
    Abort("SetDefaults(): Coarsening_StopRatio out of range") ;
  
  if ( pOptions->Coarsening_HVtxMaxFractionOfWeight <= 0 || 
       pOptions->Coarsening_HVtxMaxFractionOfWeight > 1 )
    Abort("SetDefaults(): Coarsening_HVtxMaxFractionOfWeight out of range") ;

  if ( pOptions->FM_InitPart_NrRestarts < 1 )
    Abort("SetDefaults(): FM_InitPart_NrRestarts out of range") ;
  
  if ( pOptions->FM_InitPart_MaxNrLoops < 1 )
    Abort("SetDefaults(): FM_InitPart_MaxNrLoops out of range") ;
  
  if ( pOptions->FM_InitPart_MaxNrNoGainMoves < 1 )
    Abort("SetDefaults(): FM_InitPart_MaxNrNoGainMoves out of range") ;
  
  if ( pOptions->FM_Refine_MaxNrLoops < 1 )
    Abort("SetDefaults(): FM_Refine_MaxNrLoops out of range") ;

  if ( pOptions->FM_Refine_MaxNrNoGainMoves < 1 )
    Abort("SetDefaults(): FM_Refine_MaxNrNoGainMoves out of range") ;


  /*## Set the random seed: ##*/
  if ( pOptions->Seed == -1 )
    {
#ifdef UNIX
      gettimeofday(&time, NULL) ;
      seed = (unsigned int) time.tv_usec ;
      srand(seed) ;
#endif
#ifndef UNIX
    Abort("SetDefaults(): wall clock seed can only be used on a UNIX system") ;
#endif
    }
  else
    {
      seed = (unsigned int) pOptions->Seed ;
      srand(seed) ;
    }
}

void SetOption(struct opts *pOptions, char *option, char *value)
{

  if ( ! strcmp(option, "LoadbalanceStrategy") )
    {      
      if ( ! strcmp(value, "dynamic") )
	pOptions->LoadbalanceStrategy = Dynamic ;
      else
      if ( ! strcmp(value, "increase") )
	pOptions->LoadbalanceStrategy = Increase ;
      else
      if ( ! strcmp(value, "decrease") )
	pOptions->LoadbalanceStrategy = Decrease ;
      else
	{
	  fprintf(stderr, "Unknown %s '%s'\n", option, value) ;
	  Abort("SetOption(): incorrect default value") ;
	}

      return ;
    }
  else
  if ( ! strcmp(option, "SplitStrategy") )
    {      
      if ( ! strcmp(value, "alternate") )
	pOptions->SplitStrategy = Alternate ;
      else
      if ( ! strcmp(value, "localbest") )
	pOptions->SplitStrategy = LocalBest ;
      else
      if ( ! strcmp(value, "localratio") )
	pOptions->SplitStrategy = LocalRatio ;
      else
      if ( ! strcmp(value, "onedimrow") )
	pOptions->SplitStrategy = OneDimRow ;
      else
      if ( ! strcmp(value, "onedimcol") )
	pOptions->SplitStrategy = OneDimCol ;
      else
	{
	  fprintf(stderr, "Unknown %s '%s'\n", option, value) ;
	  Abort("SetOption(): incorrect default value") ;
	}
      return ;
    }
  else
  if ( ! strcmp(option, "Alternate_FirstDirection") )
    {      
      if ( ! strcmp(value, "row") )
	pOptions->Alternate_FirstDirection = Row ;
      else
      if ( ! strcmp(value, "col") )
	pOptions->Alternate_FirstDirection = Col ;
      else
      if ( ! strcmp(value, "ratio") )
	pOptions->Alternate_FirstDirection = Ratio ;
      else
	{
	  fprintf(stderr, "Unknown %s '%s'\n", option, value) ;
	  Abort("SetOption(): incorrect default value") ;
	}
      return ;
    }
  else
  if ( ! strcmp(option, "SquareMatrix_DistributeVectorsEqual") )
    {
      if ( ! strcmp(value, "no") || ! strcmp(value, "0") )
	pOptions->SquareMatrix_DistributeVectorsEqual = EqVecNo ;
      else
      if ( ! strcmp(value, "yes") || ! strcmp(value, "1") )
	pOptions->SquareMatrix_DistributeVectorsEqual = EqVecYes ;
      else
	{
	  fprintf(stderr, "Unknown %s '%s'\n", option, value) ;
	  Abort("SetOption(): incorrect default value") ;
	}
      return ;
    }
  else
  if ( ! strcmp(option, "SquareMatrix_DistributeVectorsEqual_AddDummies") )
    {
      if ( ! strcmp(value, "no") || ! strcmp(value, "0") )
	pOptions->SquareMatrix_DistributeVectorsEqual_AddDummies = DumNo ;
      else
      if ( ! strcmp(value, "yes") || ! strcmp(value, "1") )
	pOptions->SquareMatrix_DistributeVectorsEqual_AddDummies = DumYes ;
      else
	{
	  fprintf(stderr, "Unknown %s '%s'\n", option, value) ;
	  Abort("SetOption(): incorrect default value") ;
	}
      return ;
    }
  else
  if ( ! strcmp(option, "SymmetricMatrix_UseSingleEntry") )
    {
      if ( ! strcmp(value, "no") || ! strcmp(value, "0") )
	pOptions->SymmetricMatrix_UseSingleEntry = SingleEntNo ;
      else
      if ( ! strcmp(value, "yes") || ! strcmp(value, "1") )
	pOptions->SymmetricMatrix_UseSingleEntry = SingleEntYes ;
      else
	{
	  fprintf(stderr, "Unknown %s '%s'\n", option, value) ;
	  Abort("SetOption(): incorrect default value") ;
	}
      return ;
    }
  else
  if ( ! strcmp(option, "SymmetricMatrix_SingleEntryType") )
    {
      if ( ! strcmp(value, "lower") )
	pOptions->SymmetricMatrix_SingleEntryType = ETypeLower ;
      else
      if ( ! strcmp(value, "random") )
	pOptions->SymmetricMatrix_SingleEntryType = ETypeRandom ;
      else
	{
	  fprintf(stderr, "Unknown %s '%s'\n", option, value) ;
	  Abort("SetOption(): incorrect default value") ;
	}
      return ;
    }
  else
  if ( ! strcmp(option, "Coarsening_NrVertices") )
    {
      pOptions->Coarsening_NrVertices = atoi(value) ;

      return ;
    }
  else
  if ( ! strcmp(option, "Coarsening_MaxNrVtxInMatch") )
    {
      pOptions->Coarsening_MaxNrVtxInMatch = atoi(value) ;

      return ;
    }
  else
  if ( ! strcmp(option, "Coarsening_StopRatio") )
    {
      pOptions->Coarsening_StopRatio = atof(value) ;

      return ;
    }
  else
  if ( ! strcmp(option, "Coarsening_HVtxMaxFractionOfWeight") )
    {
      pOptions->Coarsening_HVtxMaxFractionOfWeight = atof(value) ;

      return ;
    }
  else
  if ( ! strcmp(option, "Coarsening_MatchingStrategy") )
    {
      if ( ! strcmp(value, "random") )
	pOptions->Coarsening_MatchingStrategy = Random ;
      else
      if ( ! strcmp(value, "inproduct") )
	pOptions->Coarsening_MatchingStrategy = Inproduct ;
      else
	{
	  fprintf(stderr, "Unknown %s '%s'\n", option, value) ;
	  Abort("SetOption(): incorrect default value") ;
	}
      return ;
    }
  else
  if ( ! strcmp(option, "Coarsening_WeightedHyperEdges") )
    {
      if ( ! strcmp(value, "no") || ! strcmp(value, "0") )
	pOptions->Coarsening_WeightedHyperEdges = WgtEdgeNo ;
      else
      if ( ! strcmp(value, "yes") || ! strcmp(value, "1") )
	pOptions->Coarsening_WeightedHyperEdges = WgtEdgeYes ;
      else
	{
	  fprintf(stderr, "Unknown %s '%s'\n", option, value) ;
	  Abort("SetOption(): incorrect default value") ;
	}
      return ;
    }
  else
  if ( ! strcmp(option, "FM_WeightedHyperEdges") )
    {
      if ( ! strcmp(value, "no") || ! strcmp(value, "0") )
	pOptions->FM_WeightedHyperEdges = FMWgtEdgeNo ;
      else
      if ( ! strcmp(value, "yes") || ! strcmp(value, "1") )
	pOptions->FM_WeightedHyperEdges = FMWgtEdgeYes ;
      else
	{
	  fprintf(stderr, "Unknown %s '%s'\n", option, value) ;
	  Abort("SetOption(): incorrect default value") ;
	}
      return ;
    }
  else
  if ( ! strcmp(option, "FM_InitPart_NrRestarts") )
    {
      pOptions->FM_InitPart_NrRestarts = atoi(value) ;

      return ;
    }
  else
  if ( ! strcmp(option, "FM_InitPart_MaxNrLoops") )
    {
      pOptions->FM_InitPart_MaxNrLoops = atoi(value) ;

      return ;
    }
  else
  if ( ! strcmp(option, "FM_InitPart_MaxNrNoGainMoves") )
    {
      pOptions->FM_InitPart_MaxNrNoGainMoves = atoi(value) ;

      return ;
    }
  else
  if ( ! strcmp(option, "FM_Refine_MaxNrLoops") )
    {
      pOptions->FM_Refine_MaxNrLoops = atoi(value) ;

      return ;
    }
  else
  if ( ! strcmp(option, "FM_Refine_MaxNrNoGainMoves") )
    {
      pOptions->FM_Refine_MaxNrNoGainMoves = atoi(value) ;

      return ;
    }
  else
  if ( ! strcmp(option, "VectorPartition_Step3") )
    {
      if ( ! strcmp(value, "increase") )
	pOptions->VectorPartition_Step3 = Increasing3 ;
      else
      if ( ! strcmp(value, "decrease") )
	pOptions->VectorPartition_Step3 = Decreasing3 ;
      else
      if ( ! strcmp(value, "random") )
	pOptions->VectorPartition_Step3 = Random3 ;
      else
	{
	  fprintf(stderr, "Unknown %s '%s'\n", option, value) ;
	  Abort("SetOption(): incorrect default value") ;
	}

      return ;      
    }
  else
  if ( ! strcmp(option, "Seed") )
    {
      if ( ! strcmp(value, "random") )
	pOptions->Seed = -1 ;
      else
	pOptions->Seed = labs(atol(value)) ; 

      return ;      
    }
  else
    {
      fprintf(stderr, "Unknown option '%s' \n", option) ;
      Abort("SetOption(): incorrect option") ;
    }
}


void PrintHelp(int argc, char **argv) 
{
  printf("\nMondriaan version %s.\n", mversion) ;
  printf("Usage: %s [matrix] [P] [eps] <options>\n\n", argv[0]) ;
  printf(" [matrix] - the matrix to partition\n") ;
  printf(" [P]      - the number of processors\n") ;
  printf(" [eps]    - the maximum allowed load imbalance\n\n") ;
  printf(" <options>: ") ;
  printf(" (see README for details)\n") ;
  printf("\n") ;
  fflush(stdout) ;
}
